{-----------------------------------------------------------------------------
 Unit Name: uAreaDetection
 Author:    sam
 Date:      28-Nov-2021
 Purpose:   This unit contains function to detect area's in a black&white bitmap.
 History:
-----------------------------------------------------------------------------}
unit uAreaDetection;

{$IFDEF FPC}
 {$MODE Delphi}
{$ENDIF}

interface

uses
{$IFDEF FPC}
 Classes, Types, Generics.Collections, Graphics,
{$ELSE}
 System.Classes,
 System.Types,
 System.Generics.Collections,
 WinAPI.Windows,
 Vcl.Graphics,
{$ENDIF}
 uImageDefs;


const
cHeaderString='AreaNo.;CenterPoint;TopLeft;TopRight;BottomRight;BottomLeft;(Width,Height);Ratio;WhitePixelsInsideAreaPercent;NoWhiteOfPixels;FocusPoint;Distance;Propability;LightSource;AreaMembers';

type

TRegionMembers=Class(TObject)
private
  FItems:TList<Integer>;
  procedure SetItem(_Index:integer;Value:Integer);
  function  GetItem(_Index:integer):Integer;
public
  property Items[_Index:integer]:integer read GetItem write SetItem;
  function IsRegionMemberof(const _RegionNo:integer):boolean;
  procedure SortMembers;
  procedure Add(const _RegionMember:integer);
  function Count:integer;
  constructor Create;
  destructor Destroy;override;
  function AsString:string;
end;

TRegion=class(TObject)
  MinHeight: integer;
  MinWidth: integer;
  No:integer;
  Members:TRegionMembers;
  Rect:TRect;
  TopLeft:TPoint;
  WhitePixelsInsideArea:integer;        // number of pixels belonging to this area.
  WhitePixelsInsideAreaPercent:integer; // number of pixels belonging to this area in %
  CenterPoint:TPoint;                   // center point of the area
  CenterPointHasColor:boolean;          // is set to true, if the pixel at the center point is white
  CenterRect:TRect;                     // center-rect of the area. 15% of the height, 15% of the width
  FocusPoint:TPoint;
  FocusPointAbsolute:TPoint;
  FocusPointIsinCenter:boolean;
  Probability:integer;
  Distance:double;      // object distance read from exif-information (if available)
  LightSource:byte;
  constructor Create;
  destructor Destroy;override;
  procedure Calculate;
  procedure CopyFrom(const _Region:TRegion);
  function AsString:string;
  function Ratio:double;         // Ratio= with/height
end;


TRegionsList=class
  private
    FList:TList;
    FOwnsItems: boolean;
    function GetHiggestProbablityRegion: TRegion;
    function GetHiggestProbablity: integer;
    function GetStdDevY: integer;
    function GetBiggestRegion:TRegion;
  public
    constructor Create;
    destructor Destroy;override;
    procedure Add(const _Region:TRegion);
    procedure Delete(const _Index:integer);
    procedure Insert(const _Index:Integer;_Region:TRegion);
    function GetItem(const _Index:integer):TRegion;
    procedure Consolidate(const _MinWidth,_MinHeight:integer);
    procedure SaveToFile(const _filename:string);
    function GetRegionNo(const MemberNo,x,y:integer):integer;
    function Count:integer;
    function AsStringList:TStringList;
    function AsString:string;
    procedure JoinRegionsAboveEachOther(const _Max_Y_Distance_Percent,_Max_CP_X_Deviation_Percent:integer);
    procedure JoinOverlappingRegions;
    property Items[const _index:integer]:TRegion read GetItem;
    property BiggestRegion:TRegion read GetBiggestRegion;
    property HiggestProbablityRegion:TRegion read GetHiggestProbablityRegion;
    property HiggestProbablity:integer read GetHiggestProbablity;
    property OwnsItems:boolean read FOwnsItems write FOwnsItems;
    property StdDevY:integer read GetStdDevY;
end;


function DetectObjectAreas(const _Bitmap:TBitmap;                     // the image to analyze (black&white)
                           const _AreaToAnalyze:TRect;                // the area of the image to analyze. If left empty, then the whole bitmap will be analayzed
                           const _MinWidth:integer;                   // minimum width of the area we are looking for.
                           const _MinHeight:integer;                  // minimum height of the area we are looking for.
                           const _MinWhitePixelsInPercent:integer;    // minimum number of white pixels inside the detected area (in percent)
                           const _MaxWhitePixelsInPercent:integer;    // maximum number of white pixels inside the detected area (in percent)
                           const _MinRatio,_MaxRatio:double;          // min/max aspect ratio (widht/height)
                           out ObjectList:TRegionsList):boolean;      // output: List with detected areas.
                           overload;

implementation

uses
 {$IFDEF FPC}
  SysUtils, Generics.Defaults, Math,
 {$ELSE}
  System.SysUtils,
  System.Generics.Defaults,
  System.Math,
 {$ENDIF}
  uImageManipulation;

{
  |NW(X-1,Y-1)|N(X,Y-1)|NE(X+1,Y-1)|
  |W (X-1,Y   |P(X,Y)  |E (X+1,Y)  |
  |SW(X-1,Y+1 |S(X,Y+1)|SE(X+1,Y+1)|}

{-----------------------------------------------------------------------------
  Procedure: DetectObjectAreas
  Author:    sam
  Date:      23-Mai-2021
  Arguments: const
  Result:    boolean
  Description: Try to detect area's in the image <Bitmap>.
               Returns a list of identified areas in <ObjectList>.
               The areas must have at least <_MinWidth> and <_MinHeight>.
  HoshenKopelman algorithm
  https://en.wikipedia.org/wiki/Connected-component_labeling

  On the first pass:

    Iterate through each element of the data by column, then by row (Raster Scanning)  (1)
    If the element is not the background                                               (2)
        Get the neighboring elements of the current element                            (3)
        If there are no neighbors, uniquely label the current element and continue     (4)
        Otherwise, find the neighbor with the smallest label and assign it to the current element  (5)
        Store the equivalence between neighboring labels   (6)
-----------------------------------------------------------------------------}
function DetectObjectAreas(const _Bitmap:TBitmap;
                           const _AreaToAnalyze:TRect;
                           const _MinWidth:integer;
                           const _MinHeight:integer;
                           const _MinWhitePixelsInPercent:integer;
                           const _MaxWhitePixelsInPercent:integer;
                           const _MinRatio,_MaxRatio:double;
                           out ObjectList:TRegionsList):boolean;
                           overload;

const
cJoinDistance=4;
var
i:integer;
y,x: integer;
_Pixels:TPixelMap;
_SetID:integer;
_RegionList:TRegionsList;
_Region:TRegion;
_RegionLabel:byte;
_RegionNo:byte;
_NewRegion:TRegion;
_BiggestAreaSize:integer;

  function GetPixelValue(_x,_y:integer):byte;
  begin
    result:=0;
    if _x<0 then exit;
    if _x>_Bitmap.Width-1 then exit;
    if _y<0 then exit;
    if _y>_Bitmap.Height-1 then exit;
    result:=_Pixels[_y,_x].rgbtred;
  end;

  function DetectPixelValue(_x,_y:integer;out Value:Byte):boolean;
  var
  _PixelValue:byte;        // the pixel to be analyzed
// neighbour pixels around
  _NWPixelValue:byte;
  _NPixelValue:byte;
  _NEPixelValue:byte;
  _WPixelValue:byte;
  _NewPixelValue:byte;
  begin
    result:=true;
    Value:=0;
    _PixelValue:=GetPixelValue(_x,_y);
    if _PixelValue=0 then exit;           //(2)

    _NWPixelValue:=GetPixelValue(_x-1,_y-1);   //(3)
    _NPixelValue:=GetPixelValue(_x,_y-1);
    _NEPixelValue:=GetPixelValue(_x+1,_y-1);
    _WPixelValue:=GetPixelValue(_x-1,_y);

    if (_NWPixelValue=0) and
       (_NPixelValue=0) and
       (_NEPixelValue=0) and
       (_WPixelValue=0) then begin       //(4)
      if _SetID<255 then begin
        inc(_SetID);
        _Region:=TRegion.create;
        _Region.No:=_SetID;
        _Region.Members.Add(_SetID);
        _Region.WhitePixelsInsideArea:=0;
        _Region.WhitePixelsInsideAreaPercent:=0;
        _Region.FocusPointAbsolute.X:=0;
        _Region.FocusPointAbsolute.Y:=0;
        _Region.MinHeight:=_MinHeight;
        _Region.MinWidth:=_MinWidth;
        _RegionList.Add(_Region);
        Value:=_SetID;
        exit;
      end
      else begin
        result:=false;
       // showmessage('to many regions. Supports only 254 regions');
        exit;
      end;
    end;
    _NewPixelValue:=254;          //(5)
    if (_NEPixelValue>0) and (_NEPixelValue<_NewPixelValue) then _NewPixelValue:=_NEPixelValue;
    if (_WPixelValue>0)  and (_WPixelValue<_NewPixelValue)  then _NewPixelValue:=_WPixelValue;
    if (_NPixelValue>0)  and (_NPixelValue<_NewPixelValue)  then _NewPixelValue:=_NPixelValue;
    if (_NWPixelValue>0) and (_NWPixelValue<_NewPixelValue) then _NewPixelValue:=_NWPixelValue;
    if _NewPixelValue=254 then Value:=_SetID
                          else Value:=_NewPixelValue;
//(6)
    if ((_WPixelValue>0) and (_WPixelValue<255)) and (Value<>_WPixelValue) then begin
      _RegionList.GetItem(Value-1).Members.Add(_WPixelValue);
      _RegionList.GetItem(_WPixelValue-1).Members.Add(Value);
    end;

    if ((_NWPixelValue>0)  and (_NWPixelValue<255)) and (Value<>_NWPixelValue) then begin
      _RegionList.GetItem(Value-1).Members.Add(_NWPixelValue);
      _RegionList.GetItem(_NWPixelValue-1).Members.Add(Value);
    end;

    if ((_NPixelValue>0) and (_NPixelValue<255)) and (Value<>_NPixelValue) then begin
      _RegionList.GetItem(Value-1).Members.Add(_NPixelValue);
      _RegionList.GetItem(_NPixelValue-1).Members.Add(Value);
    end;

     if ((_NEPixelValue>0) and (_NEPixelValue<255)) and (Value<>_NEPixelValue) then begin
      _RegionList.GetItem(Value-1).Members.Add(_NEPixelValue);
      _RegionList.GetItem(_NEPixelValue-1).Members.Add(Value);
    end;

  end;

  procedure SetPixel(_x,_y:integer;_PixelValue:byte);
  begin
    if _x<0  then exit;
    if _y<0  then exit;

    if _x>=_Bitmap.width then exit;
    if _y>=_Bitmap.height then exit;
    if _Pixels[_y,_x].rgbtred<>_PixelValue   then _Pixels[_y,_x].rgbtred  :=_PixelValue;
    if _Pixels[_y,_x].rgbtgreen<>_PixelValue then _Pixels[_y,_x].rgbtgreen:=_PixelValue;
    if _Pixels[_y,_x].rgbtblue<>_PixelValue  then _Pixels[_y,_x].rgbtblue :=_PixelValue;
  end;

begin
  result:=false;
  ObjectList:=TRegionsList.Create;
  ObjectList.OwnsItems:=true;

  if not assigned(_Bitmap) then exit;
  if _Bitmap.Empty then exit;
  if _MinWidth=0 then exit;
  if _MinHeight=0 then exit;

  _BiggestAreaSize:=0;
  _SetID:=0;
  _RegionList:=TRegionsList.Create;
  _RegionList.OwnsItems:=true;
  try
    if not Bitmap2PixelMap(_Bitmap,true,_Pixels) then exit;
    {$ifdef debug}
    _Bitmap.SaveToFile('.\01_before_detect_areadetect.bmp');
    {$endif}

// first pass (1)
    for y:= _AreaToAnalyze.top to _AreaToAnalyze.bottom do begin
      for x:= _AreaToAnalyze.left to _AreaToAnalyze.right do begin
        if not DetectPixelValue(x,y,_RegionLabel) then continue;
        if _RegionLabel>0 then SetPixel(x,y,_RegionLabel);
      end;
    end;

    {$ifdef debug}
    _Bitmap.SaveToFile('.\02_after_firstpass_areadetect.bmp');
    {$endif}

  {$ifdef debug}
    _RegionList.SaveToFile('.\02_list_after_first_pass.csv');
  {$endif}
// second pass (6)
    _RegionList.Consolidate(_MinWidth,_MinHeight);

    for x:= _AreaToAnalyze.left+1 to _AreaToAnalyze.right-1 do begin
      for y:= _AreaToAnalyze.top+1 to _AreaToAnalyze.Bottom-1 do begin
        _RegionLabel:=GetPixelValue(x,y);
        if _RegionLabel=0 then continue;
        _RegionNo:=_RegionList.GetRegionNo(_RegionLabel,x,y);
        if _RegionNo>0 then SetPixel(x,y,255);
      end;
    end;

    for i := 0 to _RegionList.Count-1 do begin
      _Region:=_RegionList.GetItem(i);
      _Region.Calculate;
    end;

  {$ifdef debug}
    _RegionList.SaveToFile('.\05_list_after_second_pass.csv');
  {$endif}

    {$ifdef debug}
    _Bitmap.SaveToFile('.\05_after_detect_areadetect.bmp');
    {$endif}


    i := _RegionList.Count-1;
    while i>-1 do begin
      _Region:=_RegionList.GetItem(i);
      if _Region.WhitePixelsInsideArea=0 then begin
        _RegionList.Delete(i);
        dec(i);
        continue;
      end;

      if (_MinWidth>0) and (_MinHeight>0) then begin       // if some minimum specifications are given
        if _Region.Rect.Height<_MinHeight then begin       // then exclude area's which are smaller than defined.
          _RegionList.Delete(i);
          dec(i);
          continue;
        end;
        if _Region.Rect.Width<_MinWidth then begin         // then exclude area's which are smaller than defined.
          _RegionList.Delete(i);
          dec(i);
          continue;
        end;
      end;

      if ((_MinWhitePixelsInPercent<>0) and (_Region.WhitePixelsInsideAreaPercent<_MinWhitePixelsInPercent)) or
         ((_MaxWhitePixelsInPercent<>0) and (_Region.WhitePixelsInsideAreaPercent>_MaxWhitePixelsInPercent)) then begin
        _RegionList.Delete(i);
        dec(i);
        continue;
      end;

      if ((_MinRatio<>0) and ((_Region.Rect.Width/_Region.Rect.Height)<_MinRatio)) or
         ((_MaxRatio<>0) and ((_Region.Rect.Width/_Region.Rect.Height)>_MaxRatio)) then begin
        _RegionList.Delete(i);
        dec(i);
        continue;
      end;

      if (_Region.Rect.Width*_Region.Rect.height)>_BiggestAreaSize then begin
        _BiggestAreaSize:=_Region.Rect.Width*_Region.Rect.height;
      end;

      _Region.CenterPointHasColor:=(GetPixelValue(_Region.CenterPoint.X,_Region.CenterPoint.y)=255);
      _NewRegion:=TRegion.Create;
      _NewRegion.CopyFrom(_Region);
      ObjectList.Add(_NewRegion);
      dec(i);
    end;

  {$ifdef debug}
    ObjectList.SaveToFile('.\06_list_after_sort_out_by_specification.csv');
  {$endif}

  for i:=1 to ObjectList.Count do TRegion(ObjectList.Items[i-1]).No:=i;

  {$ifdef debug}
    ObjectList.SaveToFile('.\07_list.csv');
  {$endif}
    result:=true;
  finally
    _RegionList.Free;
  end;
end;

{ TRegion }

constructor TRegion.Create;
begin
  inherited;
  Members:=TRegionMembers.Create;
  WhitePixelsInsideArea:=0;
  Rect.Left:=5000;   // initialized this way by purpose.
  Rect.Right:=0;
  Rect.Top:=5000;
  Rect.Bottom:=0;
end;

destructor TRegion.Destroy;
begin
  Members.Free;
  inherited;
end;

procedure TRegion.CopyFrom(const _Region:TRegion);
var
i:integer;
begin
  No:=_Region.No;
  Rect:=_Region.Rect;
  WhitePixelsInsideArea:=_Region.WhitePixelsInsideArea;
  WhitePixelsInsideAreaPercent:=_Region.WhitePixelsInsideAreaPercent;
  CenterPointHasColor:=_Region.CenterPointHasColor;
  CenterPoint:=_Region.CenterPoint;
  CenterRect:=_Region.CenterRect;
  FocusPoint:=_Region.FocusPoint;
  FocusPointAbsolute:=_Region.FocusPointAbsolute;
  FocusPointIsinCenter:=_Region.FocusPointIsinCenter;
  Probability:=_Region.Probability;
  Distance:=_Region.Distance;
  TopLeft:=_Region.TopLeft;
  for i:=0 to _Region.Members.Count do Members.Add(_Region.Members.items[i]);
end;


{-----------------------------------------------------------------------------
  Procedure: Calculate
  Author:    sam
  Date:      09-Mr-2024
  Arguments:
  Desctiption: calculate some metrics of the region.
  Result:    None
-----------------------------------------------------------------------------}
procedure TRegion.Calculate;
var
_delta_w:double;
_delta_h:double;
_llx,_ulx:integer;
_lly,_uly:integer;
begin
  FocusPoint.X:=0;
  FocusPoint.Y:=0;
  CenterPoint.X:=0;
  CenterPoint.Y:=0;
  WhitePixelsInsideAreaPercent:=0;
  FocusPointIsinCenter:=false;
  if WhitePixelsInsideArea=0 then exit;
  if Rect.Width<MinWidth then exit;
  if Rect.Height<MinHeight then exit;


// focus point of the area
  FocusPoint.X:=round(FocusPointAbsolute.X/WhitePixelsInsideArea);
  FocusPoint.Y:=round(FocusPointAbsolute.Y/WhitePixelsInsideArea);

// center of the area
  CenterPoint.X:=Rect.Left+(Rect.Width div 2);
  CenterPoint.Y:=Rect.Top+(Rect.Height div 2);

// center square of the area
  _delta_w:= trunc(Rect.width  *0.16);
  _delta_h:= trunc(Rect.height *0.16);

  _llx:=Rect.Left + trunc((Rect.width/2) - _delta_w);
  _ulx:=Rect.Left + trunc((Rect.width/2) + _delta_w);
  _lly:=Rect.Top + trunc((Rect.Height/2) - _delta_h);
  _uly:=Rect.Top + trunc((Rect.Height/2) + _delta_h);

  CenterRect.Left:=_llx;
  CenterRect.Right:=_ulx;
  CenterRect.Top:=_lly;
  CenterRect.Bottom:=_uly;

  TopLeft.X:=_llx;
  TopLeft.Y:=_lly;

  FocusPointIsinCenter:=PtInRect(CenterRect,FocusPoint);

  WhitePixelsInsideAreaPercent:=round((WhitePixelsInsideArea / ((Rect.Width+1)*(Rect.Height+1)))*100);
end;

function TRegion.AsString:string;
var
_m:string;
begin
  result:=No.ToString+';';
  if assigned(Members) then _m:=Members.AsString
                       else _m:='';
  result:=result+format('(%d,%d);',[Rect.CenterPoint.X,Rect.CenterPoint.Y])+
      format('(%d,%d);',[Rect.Left,Rect.top])+
      format('(%d,%d);',[Rect.Right,Rect.top])+
      format('(%d,%d);',[Rect.Right,Rect.Bottom])+
      format('(%d,%d);',[Rect.Left,Rect.Bottom])+
      format('(%d,%d);',[Rect.width,Rect.height])+
      format('%f;',[Ratio])+
      format('%d;',[WhitePixelsInsideAreaPercent])+
      format('%d;',[WhitePixelsInsideArea])+
      format('(%d,%d);',[FocusPoint.X,FocusPoint.Y])+
      format('%f;',[Distance])+
      format('%d;',[Probability])+
      format('%d;',[LightSource])+
      _m;
end;

function TRegion.Ratio:double;
begin
  result:=0;
  if Rect.Width=0 then exit;
  if Rect.Height=0 then exit;
  result:= Rect.Width/Rect.Height;
end;


{ TRegionsList }

procedure TRegionsList.Add(const _Region: TRegion);
begin
  FList.Add(_Region);
end;

{-----------------------------------------------------------------------------
  Procedure: AsString
  Author:    sam
  Date:      29-Aug-2023
  Arguments: None
  Result:    string
  Description:
-----------------------------------------------------------------------------}
function TRegionsList.AsString: string;
var
_Lines:TStringList;
begin
  _Lines:=AsStringList;
  try
    result:=_Lines.Text;
  finally
    _Lines.Free;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: AsStringList
  Author:    sam
  Date:      29-Aug-2023
  Arguments: None
  Result:    TStringList
  Description:
-----------------------------------------------------------------------------}
function TRegionsList.AsStringList:TStringList;
var
i:integer;
_Region: TRegion;
begin
  result:=TStringList.Create;
  for i:=0 to FList.Count-1 do begin
    _Region:=FList[i];
    result.Add(_Region.AsString);
  end;
  result.Insert(0,cHeaderString);
end;

constructor TRegionsList.create;
begin
  FList:=TList.Create;
end;

procedure TRegionsList.Delete(const _Index: integer);
begin
  if _index<0 then exit;
  if _index>=FList.Count then exit;
  if FOwnsItems then TRegion(FList[_Index]).Free;
  FList.Delete(_Index);
end;

destructor TRegionsList.Destroy;
begin
  while FList.Count>0 do Delete(0);
  FList.Free;
  inherited;
end;

function TRegionsList.GetHiggestProbablityRegion: TRegion;
var
i:integer;
_MaxPropability:integer;
_MaxPropabilityIdx:integer;
begin
  result:=nil;
  _MaxPropabilityIdx:=-1;
  _MaxPropability:=0;
  for i :=0 to count-1 do begin
    if TRegion(Items[i]).Probability>_MaxPropability then begin
      _MaxPropability:=TRegion(Items[i]).Probability;
      _MaxPropabilityIdx:=i;
    end;

  end;
  if _MaxPropabilityIdx=-1 then exit;
  result:=FList[_MaxPropabilityIdx];
end;

function TRegionsList.GetHiggestProbablity: integer;
var
_Region:TRegion;
begin
  result:=0;
  _Region:=GetHiggestProbablityRegion;
  if not assigned(_Region) then exit;
  result:=_Region.Probability;
end;

function TRegionsList.GetItem(const _Index: integer): TRegion;
begin
  result:=FList[_Index];
end;

function TRegionsList.GetRegionNo(const MemberNo,x,y: integer): integer;
var
i:integer;
_Region: TRegion;
begin
  result:=0;
  for i := 0 to FList.Count-1 do begin
    _Region:=FList[i];
    if not _Region.Members.IsRegionMemberof(MemberNo) then continue;
    result:=_Region.Members.Items[0];
    inc(_Region.WhitePixelsInsideArea);
    _Region.FocusPointAbsolute.X:=_Region.FocusPointAbsolute.X+x;
    _Region.FocusPointAbsolute.Y:=_Region.FocusPointAbsolute.Y+y;
    if x<_Region.Rect.Left   then _Region.Rect.Left:=x;
    if x>_Region.Rect.Right  then _Region.Rect.Right:=x;
    if y<_Region.Rect.top    then _Region.Rect.top:=y;
    if y>_Region.Rect.Bottom then _Region.Rect.Bottom:=y;
    break;
  end;
end;

function TRegionsList.GetStdDevY: integer;
var
i:integer;
_Region: TRegion;
_Data:array of single;
begin
  result:=0;
  if FList.Count<2 then exit;
  SetLength(_Data,FList.Count);
  for i := 0 to FList.Count-1 do begin
    _Region:=FList[i];
    _Data[i]:=_Region.CenterPoint.Y;
  end;
  result:=trunc(stddev(_Data));
  SetLength(_Data,0);
end;

procedure TRegionsList.Insert(const _Index: Integer; _Region: TRegion);
begin
  FList.Insert(_Index,_Region);
end;

{-----------------------------------------------------------------------------
  Procedure: Consolidate
  Author:    sam
  Date:      28-Nov-2021
  Arguments: None
  Result:    None
  Description: consolidate the area-labels
-----------------------------------------------------------------------------}
procedure TRegionsList.Consolidate(const _MinWidth,_MinHeight:integer);
var
i,j,k:integer;
_Region: TRegion;
_tmpRegion:TRegion;
_RegionNoToCheck:integer;
begin
  for i := 0 to FList.Count-1 do begin
    _Region:=FList[i];
    _Region.Members.SortMembers;
  end;
  {$ifdef debug}
    SaveToFile('.\03_before_considolate_list.csv');
  {$endif}

  i:=FList.Count-1;
  while i>-1 do begin
    _Region:=FList[i];
    _RegionNoToCheck:=_Region.No;
    for j := 0 to FList.Count-1 do begin
      _tmpRegion:=FList[j];
      if _Region=_tmpRegion then continue;
      if _tmpRegion.Members.IsRegionMemberof(_RegionNoToCheck) then begin
        for k:=0 to _Region.Members.Count-1 do begin
          _tmpRegion.Members.Add(_Region.Members.Items[k]); // add regionMember to the other entry
        end;
      end;
    end;
    dec(i);
  end;

  for i := 0 to FList.Count-1 do begin
    _Region:=FList[i];
    _Region.Members.SortMembers;
  end;

  {$ifdef debug}
    SaveToFile('.\04_after_considolate_list.csv');
  {$endif}

end;

function TRegionsList.Count: integer;
begin
  result:=FList.Count;
end;

{-----------------------------------------------------------------------------
  Procedure: SaveToFile
  Author:    sam
  Date:      28-Nov-2021
  Arguments: const _filename: string
  Result:    None
  Description: helpful to save the region's to a text-file.
-----------------------------------------------------------------------------}
procedure TRegionsList.SaveToFile(const _filename: string);
var
_File:TStringList;
begin
  _File:=AsStringList;
  try
    _File.SaveToFile(_filename);
  finally
    _File.Free;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: JoinOverlappingRegions
  Author:    sam
  Date:      23-Jul-2024
  Arguments: None
  Result:    None
  Description: Join regions which overlap each other to one new region.
-----------------------------------------------------------------------------}
procedure TRegionsList.JoinOverlappingRegions;
var
i,j:integer;
_Region1:TRegion;
_Region2:TRegion;
begin
  i:=0;
  while i<FList.Count do begin
    _Region1:=FList.Items[i];
    j:=0;
    while j<FList.Count do begin
      _Region2:=FList.Items[j];

      if (_Region1<>_Region2) and
         _Region1.Rect.IntersectsWith(_Region2.Rect) then begin
        TRegion(FList.Items[i]).Rect.Union(_Region2.Rect);
        TRegion(FList.Items[i]).WhitePixelsInsideArea:=_Region1.WhitePixelsInsideArea+_Region2.WhitePixelsInsideArea;
        TRegion(FList.Items[i]).FocusPointAbsolute.X:= _Region1.FocusPointAbsolute.X+_Region2.FocusPointAbsolute.X;
        TRegion(FList.Items[i]).FocusPointAbsolute.Y:= _Region1.FocusPointAbsolute.Y+_Region2.FocusPointAbsolute.Y;
        TRegion(FList.Items[i]).Calculate;
        TRegion(FList.Items[j]).Free;
        FList.Delete(j);
        i:=0;
        break;
      end else inc(j);
    end;
    inc(i);
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: JoinRegionsAboveEachOther
  Author:    sam
  Date:      09-Mr-2024
  Arguments: const _MinWidth,_MinHeight,_Distance:integer
  Description: This unit joins two region's which have near the same width and
               are above each other.

  e.g. a face with glasses might be detected as two areas (1) & (2).
  This method will join this two areas to make one.
                ____________                 ___________
               |    1      |                |    UR    |
              |_____  _____|                |__________|
              ||   |_|   | |                 ____VD___
              ||___| |___| |                |        |
              |     2     |                 |   LR   |
              |  |____|   |                 |________|
               |_________|

  The parameter <_Max_Y_Distance_Percent> (VD) defines the maximal vertical distance between
  UR(upper Region) and LR (lower Region).
  The parameter <_Max_CP_X_Deviation_Percent> defines the maximal horizontal CP_X) deviation of
  the upper and lower region.

  Result:    None
-----------------------------------------------------------------------------}
procedure TRegionsList.JoinRegionsAboveEachOther(const _Max_Y_Distance_Percent,_Max_CP_X_Deviation_Percent:integer);
var
i,j:integer;
_UpperRegion:TRegion;
_LowerRegion:TRegion;
_Diff_CP_X:integer;    // absolute (centerpoint.x of upper-region - centerpoint.x of lower-region)
begin
  i:=0;
  while i<FList.Count do begin
    _UpperRegion:=FList.Items[i];
    j:=0;
    while j<FList.Count do begin
      _LowerRegion:=FList.Items[j];
      _Diff_CP_X:=abs(_UpperRegion.CenterPoint.x-_LowerRegion.CenterPoint.x);
      if (_UpperRegion<>_LowerRegion) and
         (_Diff_CP_X<20) and
         (_UpperRegion.Rect.Bottom-_LowerRegion.Rect.Top<20) and
         (abs(_UpperRegion.Rect.Width-_LowerRegion.Rect.Width)<25) then begin
        TRegion(FList.Items[i]).Rect.Top:=_UpperRegion.Rect.Top;
        if _LowerRegion.Rect.Left<_UpperRegion.Rect.left then TRegion(FList.Items[i]).Rect.Left:=_LowerRegion.Rect.Left
                                                         else TRegion(FList.Items[i]).Rect.Left:=_UpperRegion.Rect.Left;
        TRegion(FList.Items[i]).Rect.Bottom:=_LowerRegion.Rect.Bottom;
        if _LowerRegion.Rect.right>_UpperRegion.Rect.right then TRegion(FList.Items[i]).Rect.right:=_LowerRegion.Rect.right
                                                           else TRegion(FList.Items[i]).Rect.right:=_UpperRegion.Rect.right;

        TRegion(FList.Items[i]).WhitePixelsInsideArea:=_UpperRegion.WhitePixelsInsideArea+_LowerRegion.WhitePixelsInsideArea;
        TRegion(FList.Items[i]).FocusPointAbsolute.X:= _UpperRegion.FocusPointAbsolute.X+_LowerRegion.FocusPointAbsolute.X;
        TRegion(FList.Items[i]).FocusPointAbsolute.Y:= _UpperRegion.FocusPointAbsolute.Y+_LowerRegion.FocusPointAbsolute.Y;
        TRegion(FList.Items[i]).Calculate;
        TRegion(FList.Items[j]).Free;
        FList.Delete(j);
        i:=0;
        break;
      end else inc(j);
    end;
    inc(i);
  end;
end;

function TRegionsList.GetBiggestRegion:TRegion;
var
i: Integer;
_Region:TRegion;
_BiggestSize:integer;
begin
  result:=nil;
  _BiggestSize := 0;
  for i := 0 to FList.Count - 1 do begin // determ the biggest area available
    _Region := FList.Items[i];
    if _BiggestSize < (_Region.Rect.Width * _Region.Rect.Height) then begin
      _BiggestSize :=  _Region.Rect.Width * _Region.Rect.Height;
      result:=_Region;
    end;
  end;
end;

{ TRegionMember }

function TRegionMembers.IsRegionMemberof(const _RegionNo: integer): boolean;
var
i:integer;
begin
  result:=false;
  for i := 0 to FItems.Count-1 do begin
    if _RegionNo<>FItems[i] then continue;
    result:=true;
    exit;
  end;
end;

procedure TRegionMembers.Add(const _RegionMember: integer);
begin
  if _RegionMember=0 then exit;
  if FItems.IndexOf(_RegionMember)>-1 then exit;  // is already here
  FItems.Add(_RegionMember);
end;

function TRegionMembers.Count: integer;
begin
  result:=FItems.Count;
end;

constructor TRegionMembers.Create;
begin
  FItems:=TList<Integer>.Create;
end;

destructor TRegionMembers.Destroy;
begin
  while FItems.Count>0 do FItems.Delete(0);
  FItems.Free;
  inherited;
end;

function TRegionMembers.GetItem(_Index: integer): Integer;
begin
  result:=-1;
  if (_Index<0) or (_Index>=FItems.Count) then exit;
  result:=FItems[_Index];
end;


procedure TRegionMembers.SetItem(_Index, Value: Integer);
begin
  if (_Index<0) or (_Index>=FItems.Count) then begin
    FItems.Add(Value);
    exit;
  end;
  FItems[_Index]:=Value;
end;

procedure TRegionMembers.SortMembers;
begin
 {$IFDEF FPC}
  FItems.Sort;    // wp: FIXME - Comparer missing !!!
 {$ELSE}
  FItems.Sort(
  TComparer<Integer>.Construct(
    function(const A, B: integer): Integer
    begin
      Result := (A-B);
    end
    )
  );
 {$ENDIF}
end;

function TRegionMembers.AsString:string;
//var
//j:integer;
begin
  result:='';
//  for j:=0 to FItems.Count-1 do begin
//    if j>0 then result:=result+','+inttostr(FItems[j])    // gives crashes in some cases!!
//           else result:=inttostr(FItems[j]);              // don't know why.
//  end;
end;

function JoinOverlappingRectangles(rectangles: TList<TRect>): TList<TRect>;
var
  i, j: Integer;
  mergedRect: TRect;
  overlapFound: Boolean;
begin
  Result := TList<TRect>.Create;
  for i := 0 to rectangles.Count - 1 do begin
    overlapFound := False;
    mergedRect := rectangles[i];
    for j := 0 to Result.Count - 1 do begin
      if IntersectRect(mergedRect, mergedRect, Result[j]) then begin
        UnionRect(mergedRect, mergedRect, Result[j]);
        Result.Delete(j);
        overlapFound := True;
        Break;
      end;
    end;
    if not overlapFound then Result.Add(mergedRect);
  end;
end;

end.
