{-----------------------------------------------------------------------------
 Unit Name: uDetectFace
 Author:    sam
 Date:      24-Jul-2024
 Purpose:
 History:   try to find out if a the area could contain a face or not.
-----------------------------------------------------------------------------}
unit uDetectFace;

{$IFDEF FPC}
 {$MODE Delphi}
{$ENDIF}

interface

uses
 {$IFDEF FPC}
  SysUtils, Classes, Types, Generics.Collections, Math, Graphics, GraphType,
 {$ELSE}
  System.Classes,
  System.Types,
  System.Generics.Collections,
  WinApi.Windows,
  VCL.Graphics,
 {$ENDIF}
  uAreaDetection,
  uImageDefs;

type

TFace=class
public
  Head:TRegion;
  LeftEye:TRegion;
  RightEye:TRegion;
  Mouth:TRegion;
  Probability:integer;
  constructor create;
  destructor Destroy;override;
  function AsString:string;
end;

TFaceList=TList<TFace>;

TFaces=class(TFaceList)
private
  function GetHiggestProbablityRegion: TRegion;
  function GetHiggestProbablity: integer;
public
  function AsStringList:TStringList;
  function AsString:string;
  destructor Destroy;override;
  procedure SaveToFile(_filename:string);
  property HiggestProbablityRegion:TRegion read GetHiggestProbablityRegion;
  property HiggestProbablity:integer read GetHiggestProbablity;
end;


procedure IsFace(var Bitmap: TBitmap;_ObjectArea:TRegion;out Face:TFace);

implementation

uses
 {$IFNDEF FPC}
  System.SysUtils,
  System.Math,
 {$ENDIF}
  uImageManipulation;

{-----------------------------------------------------------------------------
  Procedure: FindMouth
  Author:    sam
  Date:      13-Feb-2024
  Arguments: const _HeadRect:TRect;const _FaceList:TRegionsList;out Mouth:TRegion
  Result:    boolean
  Description: The parameter <_HeadRect> defines the the total size of the head.
               The list <_FaceList> contains detected parts inside the head-area.
               Out-Parameter <Mouth> will contain the area with the highest
               Probability to be the "mouth".
-----------------------------------------------------------------------------}
function FindMouth(const _HeadRect:TRect;const _FaceList:TRegionsList;out Mouth:TRegion):boolean;
var
i:integer;
_Area:TRegion;
_ExpectedTargetRect:TRect;
_Candidates:TRegionsList;
begin
  result:=false;
  Mouth:=nil;
  i:=0;
  if _FaceList.Count=0 then exit;
  _Candidates:=TRegionsList.create;
  try
    _ExpectedTargetRect.Left :=trunc(_HeadRect.Width*0.05);
    _ExpectedTargetRect.Right:=trunc(_HeadRect.Width*0.95);
    _ExpectedTargetRect.Top:=trunc((_HeadRect.Height/2)*1.02);
    _ExpectedTargetRect.Bottom:=trunc(_HeadRect.Height*0.95);
    while i<_FaceList.count do begin
      _Area:=_FaceList.GetItem(i);
      if (_Area.Rect.Left>_ExpectedTargetRect.Left) and
         (_Area.Rect.Right<_ExpectedTargetRect.Right) and
         (_Area.Rect.Top>_ExpectedTargetRect.Top) and
         (_Area.Rect.Bottom<_ExpectedTargetRect.Bottom) then begin
           _Candidates.Add(_Area);
         end;
      inc(i);
    end;

    if (_Candidates.Count>0) and (_Candidates.Count<6) then begin
      result:=true;
      Mouth:=_Candidates.BiggestRegion;
    end;
  finally
    _Candidates.Free;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: FindLeftEye
  Author:    sam
  Date:      12-Mr-2024
  Arguments: const _HeadRect:TRect;const _FaceList:TRegionsList;const _CenterPointMouth:TPoint;out LeftEye:TRegion
  Result:    boolean
  Description: Is there an area inside the expected region for the left eye?
-----------------------------------------------------------------------------}
function FindLeftEye(const _HeadRect:TRect;const _FaceList:TRegionsList;const _CenterPointMouth:TPoint;out LeftEye:TRegion):boolean;
var
i:integer;
_Area:TRegion;
_ExpectedTargetRect:TRect;
begin
  result:=false;
  i:=0;
  _ExpectedTargetRect.Left :=0;
  if _CenterPointMouth.X>0 then _ExpectedTargetRect.Right:=trunc(_CenterPointMouth.x*0.95)
                           else _ExpectedTargetRect.Right:=trunc(_HeadRect.Width*0.6);
  _ExpectedTargetRect.Top:=0;
  if _CenterPointMouth.Y>0 then _ExpectedTargetRect.Bottom:=trunc((_CenterPointMouth.y)*0.8)
                           else _ExpectedTargetRect.Bottom:=trunc(_HeadRect.Height * 0.75);

  while i<_FaceList.count do begin
    _Area:=_FaceList.GetItem(i);
    if (_Area.Rect.Left>=_ExpectedTargetRect.Left) and
       (_Area.Rect.Right<=_ExpectedTargetRect.Right) and
       (_Area.Rect.Top>=_ExpectedTargetRect.Top) and
       (_Area.Rect.Bottom<=_ExpectedTargetRect.Bottom) then begin
         LeftEye:=_Area;
         result:=true;
         exit;
       end;
    inc(i);
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: FindRightEye
  Author:    sam
  Date:      12-Mr-2024
  Arguments: const _HeadRect:TRect;const _FaceList:TRegionsList;const _CenterPointMouth:TPoint;out RightEye:TRegion
  Result:    boolean
  Description: Is there an area inside the expected region for the right eye?
-----------------------------------------------------------------------------}
function FindRightEye(const _HeadRect:TRect;const _FaceList:TRegionsList;const _CenterPointMouth:TPoint;out RightEye:TRegion):boolean;
var
i:integer;
_Area:TRegion;
_ExpectedTargetRect:TRect;
begin
  result:=false;
  i:=0;
  if _CenterPointMouth.X>0 then _ExpectedTargetRect.Left :=trunc(_CenterPointMouth.x*1.05)
                           else _ExpectedTargetRect.Left :=trunc(_HeadRect.Width*0.6);
  _ExpectedTargetRect.Right:=_HeadRect.Width;
  _ExpectedTargetRect.Top:=0;
  if _CenterPointMouth.Y>0 then _ExpectedTargetRect.Bottom:=trunc(_CenterPointMouth.y*0.8)
                           else _ExpectedTargetRect.Bottom:=trunc(_HeadRect.Height * 0.75);

  while i<_FaceList.count do begin
    _Area:=_FaceList.GetItem(i);
    if (_Area.Rect.Left>=_ExpectedTargetRect.Left) and
       (_Area.Rect.Right<=_ExpectedTargetRect.Right) and
       (_Area.Rect.Top>=_ExpectedTargetRect.Top) and
       (_Area.Rect.Bottom<=_ExpectedTargetRect.Bottom) then begin
         RightEye:=_Area;
         result:=true;
         exit;
       end;
    inc(i);
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: IsFace
  Author:    sam
  Date:      19-Okt-2023
  Arguments:
  Result:    None
  Description: calculate the Probability of a face inside the area <_ObjectArea>.
  The out-parameter <Probability> is defined as follows:
  0 = more than 7 areas detected.
  5 = an area detected, which could be the left or right eye or the mouth.
  10 = two areas which could be left and right eye
  15 = three areas detected, which could be eye's and mouth.
-----------------------------------------------------------------------------}
procedure IsFace(var Bitmap: TBitmap;_ObjectArea:TRegion;out Face:TFace);
const
cDistanceToBorder=2;
var
_BMPWork: TBitmap;
{$ifdef debug}_BMPPaint: TBitmap;{$endif}
_FaceList: TRegionsList;
_WorkRect: TRect;
_FaceTargetRect: TRect;
_MouthCenterPoint:TPoint;
_DebugPath:string;
{$ifdef debug}_ExpectedTargetRect:TRect;{$endif}
begin
  Face:=TFace.create;
  Face.Head:=TRegion.Create;
  Face.Head.CopyFrom(_ObjectArea);
  Face.Mouth:=nil;
  Face.LeftEye:=nil;
  Face.RightEye:=nil;
  Face.Probability:=0;
  _DebugPath:=extractfilepath(paramstr(0))+'temp\';
  _BMPWork  := TBitmap.Create;
  {$ifdef debug}_BMPPaint := TBitmap.Create;{$endif}
  try
    _BMPWork.Assign(Bitmap);
    if _ObjectArea.Rect.Height>(1.65*_ObjectArea.Rect.Width) then _ObjectArea.Rect.Height:=trunc(1.65*_ObjectArea.Rect.Width);

    _ObjectArea.Rect.Left:=trunc(0.99*_ObjectArea.Rect.Left);
    _ObjectArea.Rect.Right:=trunc(1.01*_ObjectArea.Rect.Right);
    _ObjectArea.Rect.Top:=trunc(0.99*_ObjectArea.Rect.Top);
    _ObjectArea.Rect.Bottom:=trunc(1.01*_ObjectArea.Rect.Bottom);
{$ifdef debug}_BMPWork.SaveToFile(_DebugPath+format('face\%d_01_face_input.bmp',[_ObjectArea.No]));{$endif}
    Crop(_BMPWork, _ObjectArea.Rect);
{$ifdef debug}_BMPWork.SaveToFile(_DebugPath+format('face\%d_01_face_input_cropped.bmp',[_ObjectArea.No]));{$endif}
    _WorkRect.Left := 0;
    _WorkRect.Top := 0;
    _WorkRect.Right  := _BMPWork.Width;
    _WorkRect.Bottom := _BMPWork.Height;

    ManipulateImage(_BMPWork, tgs_invert);

    _BMPWork.Canvas.Brush.Style := bsSolid;
    _BMPWork.Canvas.Brush.Color := clBlack;
    _BMPWork.Canvas.FloodFill(cDistanceToBorder, cDistanceToBorder, clBlack, fsBorder);
    _BMPWork.Canvas.FloodFill(_BMPWork.Width - cDistanceToBorder, cDistanceToBorder, clBlack, fsBorder);
    _BMPWork.Canvas.FloodFill(cDistanceToBorder, _BMPWork.Height - cDistanceToBorder, clBlack, fsBorder);
    _BMPWork.Canvas.FloodFill(_BMPWork.Width - cDistanceToBorder, _BMPWork.Height - cDistanceToBorder, clBlack, fsBorder);
{$ifdef debug}_BMPWork.SaveToFile(_DebugPath+format('face\%d_02_face_floodfill_invert.bmp',[_ObjectArea.No]));{$endif}
    RemoveSmallObjects(_BMPWork, _WorkRect,trunc(_WorkRect.Width*0.05),trunc(_WorkRect.Height*0.05), 0, _FaceTargetRect);
{$ifdef debug}_BMPWork.SaveToFile(_DebugPath+format('face\%d_03_removesmallobjects.bmp',[_ObjectArea.No]));{$endif}
    try
      DetectObjectAreas(_BMPWork,_WorkRect,2,1,10,100,0.1,5,_FaceList);
{$ifdef debug}
      _BMPPaint.Assign(_BMPWork);
      DrawAreas(_BMPPaint, _FaceList, tdr_AreaBox);
      _FaceList.SaveToFile(_DebugPath+format('face\%d_04_Detected_Areas.csv',[_ObjectArea.No]));
      _BMPPaint.SaveToFile(_DebugPath+format('face\%d_04_face_with_detected_areas.bmp',[_ObjectArea.No]));
{$endif}
// try to find mouth
   {$ifdef debug}
    _ExpectedTargetRect.Left :=trunc(_WorkRect.Width*0.05);
    _ExpectedTargetRect.Right:=trunc(_WorkRect.Width*0.95);
    _ExpectedTargetRect.Top:=trunc((_WorkRect.Height/2)*1.02);
    _ExpectedTargetRect.Bottom:=trunc(_WorkRect.Height*0.95);
   _BMPPaint.Assign(_BMPWork);
   DrawRect(_BMPPaint,_ExpectedTargetRect,clRed,4);
   {$endif}
   if FindMouth(_WorkRect,_FaceList,Face.Mouth) then begin
   {$ifdef debug}DrawRect(_BMPPaint,Face.Mouth.Rect,clBlue,4);{$endif}
   end; // else exit;
   {$ifdef debug}_BMPPaint.SaveToFile(_DebugPath+format('face\%d_04a_face_with_detected_areas_mouth.bmp',[_ObjectArea.No]));{$endif}
    if assigned(Face.Mouth) then _MouthCenterPoint:=Face.Mouth.CenterPoint
    else begin
      _MouthCenterPoint.x:=0;
      _MouthCenterPoint.y:=0;
    end;
// try to find left eye
  {$ifdef debug}
  _ExpectedTargetRect.Left :=0;
  if _MouthCenterPoint.X>0 then _ExpectedTargetRect.Right:=trunc(_MouthCenterPoint.x*0.95)
                           else _ExpectedTargetRect.Right:=trunc(_WorkRect.Width*0.6);
  _ExpectedTargetRect.Top:=0;
  if _MouthCenterPoint.Y>0 then _ExpectedTargetRect.Bottom:=trunc((_MouthCenterPoint.y)*0.8)
                           else _ExpectedTargetRect.Bottom:=trunc(_WorkRect.Height * 0.75);
  _BMPPaint.Assign(_BMPWork);
  DrawRect(_BMPPaint,_ExpectedTargetRect,clRed,4);
  {$endif}
  if FindLeftEye(_WorkRect,_FaceList,_MouthCenterPoint,Face.LeftEye) then begin
  {$ifdef debug}DrawRect(_BMPPaint,Face.LeftEye.Rect,clGreen,4);{$endif}
  end;
  {$ifdef debug}_BMPPaint.SaveToFile(_DebugPath+format('face\%d_04a_face_with_detected_areas_leftEye.bmp',[_ObjectArea.No]));{$endif}

// try to find right eye
 {$ifdef debug}
  if _MouthCenterPoint.X>0 then _ExpectedTargetRect.Left :=trunc(_MouthCenterPoint.x*1.05)
                           else _ExpectedTargetRect.Left :=trunc(_WorkRect.Width*0.6);
  _ExpectedTargetRect.Right:=_WorkRect.Width;
  _ExpectedTargetRect.Top:=0;
  if _MouthCenterPoint.Y>0 then _ExpectedTargetRect.Bottom:=trunc(_MouthCenterPoint.y*0.8)
                           else _ExpectedTargetRect.Bottom:=trunc(_WorkRect.Height * 0.75);

    _BMPPaint.Assign(_BMPWork);
    DrawRect(_BMPPaint,_ExpectedTargetRect,clRed,4);
  {$endif}
    if FindRightEye(_WorkRect,_FaceList,_MouthCenterPoint,Face.RightEye) then begin
   {$ifdef debug}DrawRect(_BMPPaint,Face.RightEye.Rect,clGreen,4);{$endif}
    end;

   {$ifdef debug}_BMPPaint.SaveToFile(_DebugPath+format('face\%d_04a_face_with_detected_areas_rightEye.bmp',[_ObjectArea.No]));{$endif}

  {$ifdef debug}
      if assigned(Face.LeftEye) or
         assigned(Face.RightEye) or
         assigned(Face.Mouth) then begin
        _BMPPaint.Assign(_BMPWork);
        if assigned(Face.LeftEye)  then DrawRect(_BMPPaint,Face.LeftEye.Rect,clGreen,2);
        if assigned(Face.RightEye) then DrawRect(_BMPPaint,Face.RightEye.Rect,clGreen,2);
        if assigned(Face.Mouth)    then DrawRect(_BMPPaint,Face.Mouth.Rect,clBlue,2);
        _BMPPaint.SaveToFile(_DebugPath+format('face\%d_05_face_with_detected_areas.bmp',[_ObjectArea.No]));
      end;
  {$endif}
      if (_FaceList.Count>0) and (_FaceList.Count<9) then begin
        if assigned(Face.LeftEye)  then Face.Probability:=Face.Probability+5;
        if assigned(Face.RightEye) then Face.Probability:=Face.Probability+5;
        if assigned(Face.Mouth)    then Face.Probability:=Face.Probability+5;
        if assigned(Face.LeftEye) and assigned(Face.RightEye) then Face.Probability:=Face.Probability+10;
        if assigned(Face.LeftEye) and
           assigned(Face.RightEye) and
           assigned(Face.Mouth) then Face.Probability:=Face.Probability+15;
      end;
    finally
      _FaceList.Free;
    end;
  finally
    _BMPWork.Free;
      {$ifdef debug}if assigned(_BMPPaint) then _BMPPaint.Free;{$endif}
  end;
end;

function TFaces.AsStringList:TStringList;
var
i:integer;
_Face:TFace;
_header:string;
begin
  result:=TStringList.Create;
  _header:='Object-Type;'+cHeaderString;
  for i:=0 to Count-1 do begin
    _Face:=Items[i];
    result.add(_Face.AsString);
  end;
  result.Insert(0,_header);
end;

function TFaces.AsString:string;
var
_Lines:TStringList;
begin
  _Lines:=AsStringList;
  try
    result:=_Lines.Text;
  finally
    _Lines.Free;
  end;
end;

procedure TFaces.SaveToFile(_filename:string);
var
_Lines:TStringList;
begin
  _Lines:=AsStringList;
  try
    _Lines.SaveToFile(_filename);
  finally
    _Lines.free;
  end;
end;

function TFaces.GetHiggestProbablityRegion: TRegion;
var
i:integer;
_MaxPropability:integer;
_MaxPropabilityIdx:integer;
begin
  result:=nil;
  _MaxPropabilityIdx:=-1;
  _MaxPropability:=0;
  for i :=0 to count-1 do begin
    if TFace(Items[i]).Head.Probability>_MaxPropability then begin
      _MaxPropability:=TRegion(Items[i]).Probability;
      _MaxPropabilityIdx:=i;
    end;

  end;
  if _MaxPropabilityIdx=-1 then exit;
  result:=TFace(Items[_MaxPropabilityIdx]).Head;
end;

destructor TFaces.Destroy;
begin
  while Count>0 do begin
    TFace(Items[0]).Free;
    Delete(0);
  end;
  inherited;
end;

function TFaces.GetHiggestProbablity: integer;
var
_Region:TRegion;
begin
  result:=0;
  _Region:=GetHiggestProbablityRegion;
  if not assigned(_Region) then exit;
  result:=_Region.Probability;
end;


{ TFace }
function TFace.AsString:string;
begin
  result:='';
  if assigned(Head) then result:=result+'Head;'+Head.AsString+#13+#10
                    else result:=result+'Head; n/a;'+#13+#10;
  if assigned(LeftEye) then result:=result+'Left eye;'+LeftEye.AsString+#13+#10
                       else result:=result+'Left eye;n/a;'+#13+#10;
  if assigned(RightEye) then result:=result+'Right eye;'+RightEye.AsString+#13+#10
                        else result:=result+'Right eye;n/a;'+#13+#10;
  if assigned(Mouth) then result:=result+'Mouth;'+Mouth.AsString+#13+#10
                     else result:=result+'Mouth;n/a;'+#13+#10;
end;

constructor TFace.Create;
begin
  Head:=nil;
  LeftEye:=nil;
  RightEye:=nil;
  Mouth:=nil;
end;

destructor TFace.Destroy;
begin
  if assigned(Head) then Head.Free;
  inherited;
end;

end.
