{-----------------------------------------------------------------------------
 Unit Name: uDetectHead
 Author:    sam
 Date:      08-Apr-2023
 Purpose:   Try to find out, if at least one head is on a picture.
 Licence:   If you use this code, you must add a reference to
            my library to the about-dialog of your app. Thanks.
 History:
-----------------------------------------------------------------------------}
unit uDetectHead;

interface

uses
  {$IFDEF FPC}
  Types, LCLType, Graphics,
  {$ELSE}
  System.Types,
  WinApi.Windows,
  VCL.Graphics,
  {$ENDIF}
  uExifInfoReader,
  uAreaDetection,
  uImageDefs,
  uDetectFace;


function IsFaceOnImage( const _Filename:string;                  // the image to be anaylzed (supports .bmp or .jpg files)
                        out OriginalBitmap:TBitmap; // the original bitmap (not resized)
                        out OutputBitmap:TBitmap;   // bitmap used for analysis
                        out Areas:TRegionsList;                  // detected areas
                        out Faces:TFaces;                        // areas which could contain a face.
                        out AnalysisText:string;       // some information about the analysis process
                        out ExifData:TExif):boolean;overload;

function IsFaceOnImage( const _Filename:string;               // the image to be anaylzed (supports .jpg files)
                        out Faces:TFaces):boolean;overload;    // areas which might contain a face.

implementation

uses
 {$IFDEF FPC}
  SysUtils, Math,
 {$ELSE}
  System.SysUtils,
  System.Math,
 {$ENDIF}
  uImageManipulation;

const
cImageSize=650;
cFaceMinProbability=80;
cMinNumberOfSkinPixeslPecent=0.5;  // the picture should have at least this number of "skin" pixels in percent.
cMaxNumberOfSkinPixeslPecent=50;   // the picture should not have more than this number of "skin" pixels in percent.
cSmallObjectsToRemove=7;
cMinWhitePixelsPercent=45;
cMaxWhitePixelsPercent=85;


cMaxFaceRatio= 1.2;               // max. widht/height of the face area.

cMaxObjectDistance=15;

{-----------------------------------------------------------------------------
  Procedure: KeepSkinColors
  Author:    sam
  Date:      23-Mai-2021
  Arguments: var Bitmap:TBitmap;const _threshold:Byte
  Result:    None
  Decription: try to remove all colors which are not skin-colors.
              Returns % of pixels with "skin-color" in out-param <AmountOfSkinColorPercent>
              of the detected "Skin-Area".
-----------------------------------------------------------------------------}
function KeepSkinColors(var Bitmap:TBitmap;
                        const _Exif:TExif;
                        out SkinArea:TRect;
                        out AmountOfSkinColorPercent:double):boolean;
  var
  x,y: integer;
  _Pixels:TPixelMap;
  _BackgroundValue:byte;
  _ForegroundValue:byte;
  _CountSkinColorPixels:integer;
  _WorkArea:TRect;

  function IsSkinColor(const _Color:TRGBTriple): boolean;
  var
  _GP:double;
  _BP:double;
  _diff:double;
  begin
    result:=false;
    if _Color.rgbtRed<_BackgroundValue then exit;
    if _Color.rgbtRed<_Color.rgbtGreen+25 then exit;
    if _Color.rgbtRed<_Color.rgbtBlue+25 then exit;
    _GP:= _Color.rgbtGreen/_Color.rgbtRed;
    _BP:= _Color.rgbtBlue/_Color.rgbtRed;
    _diff:=_GP-_BP;
    if _diff>0.23 then exit;   // remove brown colors
    if (_diff<=-0.02) and (_diff<(_ForegroundValue*0.0001)) then exit;
    if (1-_GP>0.50) or (1-_GP<0.1) then exit;
    result:=true;
  end;

  procedure ManipulateSkinPixel(const _x,_y:integer);
  begin
    if IsSkinColor(_Pixels[_y,_x]) then begin
      if _x<SkinArea.Left   then SkinArea.Left :=_x;
      if _x>SkinArea.Right  then SkinArea.Right:=_x;
      if _y<SkinArea.Top    then SkinArea.Top  :=_y;
      if _y>SkinArea.Bottom then SkinArea.Bottom:=_y;
      inc(_CountSkinColorPixels);
    end
    else begin
      if _Pixels[_y,_x].rgbtred>0   then _Pixels[_y,_x].rgbtred:=0;
      if _Pixels[_y,_x].rgbtgreen>0 then _Pixels[_y,_x].rgbtgreen:=0;
      if _Pixels[_y,_x].rgbtblue>0  then _Pixels[_y,_x].rgbtblue:=0;
    end;
  end;


begin
  result:=false;
  AmountOfSkinColorPercent:=0;
  _CountSkinColorPixels:=0;
  _WorkArea.Left:=0;
  _WorkArea.Right:=Bitmap.Width;
  _WorkArea.Top:=0;
  _WorkArea.Bottom:=Bitmap.Height;

  SkinArea.Left:=0;
  SkinArea.Top:=0;
  SkinArea.Right:=0;
  SkinArea.Bottom:=0;
  DetermBackgroundValue(Bitmap,_Exif,_BackgroundValue);
  DetermForegroundValue(Bitmap,_ForegroundValue);

  if not Bitmap2PixelMap(Bitmap,false,_Pixels) then exit;
  try
    SkinArea.Left:=65000;
    SkinArea.Top:=65000;
    SkinArea.Right:=-1;
    SkinArea.Bottom:=-1;

    for y:= 0 to Bitmap.Height-1 do begin
      for x:= 0 to Bitmap.Width-1 do ManipulateSkinPixel(x,y);
    end;

    if _CountSkinColorPixels=0 then exit;
    AmountOfSkinColorPercent:=(_CountSkinColorPixels/(_WorkArea.width*_WorkArea.height))*100;
    result:=true;
  finally
    PixelMap2Bitmap(_Pixels,Bitmap);
    SetLength(_Pixels,0);
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: FindFaceOnImage
  Author:    sam
  Date:
  Arguments:
  Result:
  Description: try to find out if face's are on the image.
  The out-param <Faces> is a list of areas which could be a face.
-----------------------------------------------------------------------------}
procedure FindFaceOnImage(_Bitmap:TBitmap;
                       const _Exif:TExif;
                       const _MinProbability:integer;
                       const _MinWhitePixelsInPercent,_MaxWhitePixelsInPercent:integer;
                       const _MinRatio,_MaxRatio:double;
                       const _SmallObjectsToRemove:integer;
                       const _GenerateAnalysisText:boolean;
                       out Areas:TRegionsList;
                       out Faces:TFaces;
                       out AnalysisText:string);
var
i:integer;
_ObjectArea:TRegion;
_SkinArea:TRect;
_SkinAreaAfterRemoveSmallObjects:TRect;
_AmountOfSkinColorPercent:double;
_TargetArea:TRect;
_Face:TFace;
_Probability:integer;
{$REGION 'debug_stuff'}
  {$ifdef debug}
  _OriginalBitmap:Vcl.Graphics.TBitmap;
  _DebugPath:string;
  _DebugBitmap:Vcl.Graphics.TBitmap;
  {$endif}
{$ENDREGION}
begin
  AnalysisText:='';
  Faces:=TFaces.Create;
  Areas:=TRegionsList.Create;
  Areas.OwnsItems:=true;
  if not assigned(_Bitmap) then exit;
  if not assigned(_Exif) then exit;
  if (_Exif.Valid) and (_Exif.ObjectDistance>8) then exit;    // don't analyse pictures where the object is to far away.

  if _GenerateAnalysisText then AnalysisText:=_Exif.AsString+#13+#10;

{$REGION 'debug_stuff'}
    {$ifdef debug}
    _OriginalBitmap:=Vcl.Graphics.TBitmap.Create;
    _OriginalBitmap.Assign(_Bitmap);
    _DebugPath:=extractfilepath(paramstr(0))+'temp\';
    _Bitmap.SaveToFile(_DebugPath+'skin\01_original_resized.bmp');
    {$endif}
{$ENDREGION}
//step01: set all pixesl which are not skin-color to black.
  KeepSkinColors(_Bitmap,_Exif,_SkinArea,_AmountOfSkinColorPercent);
  {$ifdef debug}_Bitmap.SaveToFile(_DebugPath+'skin\02_skin_onlycolors.bmp');{$endif}

  if _AmountOfSkinColorPercent<cMinNumberOfSkinPixeslPecent then begin
    if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Picture.rule01: has <%f> percent skin-pixels which is lower than the minimal number of skin-pixels <%f>.',[_AmountOfSkinColorPercent,cMinNumberOfSkinPixeslPecent])+#13+#10;
    exit;
  end;

  if _AmountOfSkinColorPercent>cMaxNumberOfSkinPixeslPecent then begin
    if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Picture.rule01: has <%f> percent skin-pixels which is higher than the maximal number of skin-pixels <%d>.',[_AmountOfSkinColorPercent,cMaxNumberOfSkinPixeslPecent])+#13+#10;
    exit;
  end;
// step02: make image black&white
  AutoBlackWhiteImage(_Bitmap,_Exif,_SkinArea);
  {$ifdef debug}_Bitmap.SaveToFile(_DebugPath+'skin\03_skin_blackandwhite.bmp');{$endif}

  if (_SkinArea.Width<_Exif.MinExpectedFaceWidth) or (_SkinArea.Height<_Exif.MinExpectedFaceHeight) then begin
    if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Picture.rule02: skin-area is smaller than minimal size w=%d,H=%d.',[_Exif.MinExpectedFaceWidth,_Exif.MinExpectedFaceHeight])+#13+#10;
    exit;
  end;
 //step03: remove small objects.
  RemoveSmallObjects(_Bitmap,_SkinArea,_SmallObjectsToRemove,_SmallObjectsToRemove,0,_SkinAreaAfterRemoveSmallObjects);
{$REGION 'debug_stuff'}
    {$ifdef debug}
    _DebugBitmap:=Vcl.Graphics.TBitmap.Create;
    _DebugBitmap.Assign(_Bitmap);
//    DrawRect(_DebugBitmap,_SkinAreaAfterRemoveSmallObjects,clRed,3);
    _DebugBitmap.SaveToFile(_DebugPath+'skin\04_skin_removesmallobjects.bmp');
    _DebugBitmap.Free;
    {$endif}
{$ENDREGION}
  Areas.Free;
 //step04: detect areas
  if not DetectObjectAreas(_Bitmap,
                           _SkinAreaAfterRemoveSmallObjects,
                           _Exif.MinExpectedFaceWidth,
                           _Exif.MinExpectedFaceHeight,
                           _MinWhitePixelsInPercent,
                           _MaxWhitePixelsInPercent,
                           _MinRatio,
                           _MaxRatio,
                           Areas) then exit;

{$REGION 'debug_stuff'}
    {$ifdef debug}
    Areas.SaveToFile(_DebugPath+'skin\05_skin_detected_areas.csv');
    _DebugBitmap:=Vcl.Graphics.TBitmap.Create;
    _DebugBitmap.Assign(_Bitmap);
    DrawAreas(_DebugBitmap,Areas,tdr_AreaBox);
    _DebugBitmap.SaveToFile(_DebugPath+'skin\05_skin_detected_areas.bmp');
    _DebugBitmap.Free;
    {$endif}
{$ENDREGION}

  Areas.JoinRegionsAboveEachOther(20,20);
  Areas.JoinOverlappingRegions;

{$REGION 'debug_stuff'}
    {$ifdef debug}
    Areas.SaveToFile(_DebugPath+'skin\05_skin_detected_and_joined_areas.csv');
    _DebugBitmap:=Vcl.Graphics.TBitmap.Create;
    _DebugBitmap.Assign(_Bitmap);
    DrawAreas(_DebugBitmap,Areas,tdr_AreaBox);
    _DebugBitmap.SaveToFile(_DebugPath+'skin\05_skin_detected_and_joined_areas.bmp');
    _DebugBitmap.Free;
    {$endif}
{$ENDREGION}

  _TargetArea.left  :=round(0.05*_Bitmap.Width);
  _TargetArea.Right :=round(0.95*_Bitmap.Width);
  _TargetArea.Top   :=round(0.05*_Bitmap.Height);
  _TargetArea.Bottom:=round(0.95*_Bitmap.Height);

  for i:=0 to Areas.Count-1 do begin    // iterate through all areas and calculate Probability
    _ObjectArea:=Areas.Items[i];
    _ObjectArea.Distance:=_Exif.ObjectDistance;
    if _GenerateAnalysisText then AnalysisText:=AnalysisText+'***********************************************************************'+#13+#10;
    if _Exif.LightSource<256 then _ObjectArea.LightSource:=_Exif.LightSource;
    _Probability:=0;

 //rule 01:  is the detected object size bigger than _MinHeight*_MinWidth
    if (_ObjectArea.Rect.Height>_Exif.MinExpectedFaceHeight) and
       (_ObjectArea.Rect.Width>_Exif.MinExpectedFaceWidth) then begin
      _Probability:=_Probability+15;
      if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Object <%d>.rule01: has the correct minimal size (w=%d,h=%d). (15)',[_ObjectArea.No,_ObjectArea.Rect.Width,_ObjectArea.Rect.Height])+#13+#10;
    end
    else AnalysisText:=AnalysisText+format('Object <%d>.rule01: has NOT the correct minimal expected size (w=%d,h=%d).',[_ObjectArea.No,_Exif.MinExpectedFaceWidth,_Exif.MinExpectedFaceHeight])+#13+#10;

  //rule 02:  is the detected area in the expected shape?

    if (_ObjectArea.Ratio>_MinRatio) and
       (_ObjectArea.Ratio<_MaxRatio) then begin
      _Probability:=_Probability+20;
      if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Object <%d>.rule02: has the correct ratio <%f>. (20)',[_ObjectArea.No,_ObjectArea.Ratio])+#13+#10;
    end
    else AnalysisText:=AnalysisText+format('Object <%d>.rule02: has NOT the correct ratio <%f>.',[_ObjectArea.No,_ObjectArea.Ratio])+#13+#10;

  //rule 03: is focus point inside center-square?
    if _ObjectArea.FocusPointIsinCenter then begin
      _Probability:=_Probability+25;
      if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Object <%d>.rule03: has focus-point inside the center-square. (25)',[_ObjectArea.No])+#13+#10;
    end
    else AnalysisText:=AnalysisText+format('Object <%d>.rule03: has NOT focus-point inside the center-square.',[_ObjectArea.No])+#13+#10;

// rule 04: is amount of white pixels inside the detected area  within this limit?
    if (_ObjectArea.WhitePixelsInsideAreaPercent>_MinWhitePixelsInPercent) and
       (_ObjectArea.WhitePixelsInsideAreaPercent<_MaxWhitePixelsInPercent)
    then begin
      _Probability:=_Probability+15;
      if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Object <%d>.rule04: has <%d>%% white Pixels. (15)',[_ObjectArea.No,_ObjectArea.WhitePixelsInsideAreaPercent])+#13+#10;
    end
    else AnalysisText:=AnalysisText+format('Object <%d>.rule04: has NOT enough <%d>%% white Pixels.',[_ObjectArea.No,_ObjectArea.WhitePixelsInsideAreaPercent])+#13+#10;
// analyze the area and try to find eye's and mouth
    IsFace(_Bitmap,_ObjectArea,_Face);

    _Probability:=(_Probability+_Face.Probability);
    if _Probability>100 then _Probability:=100;

    if _TargetArea.Contains(_ObjectArea.CenterPoint) then begin
      if _GenerateAnalysisText then AnalysisText:=AnalysisText+format('Object <%d>.Calculated Probability: %d.',[_ObjectArea.No,_Probability])+#13+#10;
      if _Probability>=_MinProbability then begin
        _ObjectArea.Probability:=_Probability;
        _Face.Probability:=_Probability;
        _Face.Head.Probability:=_Probability;
        Faces.Add(_Face);
      end
      else _Face.Free;
    end else _Face.Free;
  end;

{$REGION debug_stuff'}
    {$ifdef debug}
    Faces.SaveToFile(_DebugPath+'skin\06_detected_Faces.csv');
    _DebugBitmap:=Vcl.Graphics.TBitmap.Create;
    _DebugBitmap.Assign(_OriginalBitmap);
    DrawAreas(_DebugBitmap,Faces,tdr_AreaBox);
    _DebugBitmap.SaveToFile(_DebugPath+'skin\06_detected_Faces.bmp');
    _DebugBitmap.Free;
    _OriginalBitmap.Free;
    {$endif}
{$ENDREGION}
end;


{-----------------------------------------------------------------------------
  Procedure: IsFaceOnImage
  Author:    sam
  Date:      31-Jul-2024
  Arguments:
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function IsFaceOnImage( const _Filename:string;               // the image to be anaylzed (supports .bmp or .jpg files)
                        out OriginalBitmap:TBitmap; // the original bitmap (not resized)
                        out OutputBitmap:TBitmap;  // the bitmap used for analysis
                        out Areas:TRegionsList;               // detected areas
                        out Faces:TFaces;                     // areas which might contain a face.
                        out AnalysisText:string;
                        out ExifData:TExif):boolean;     // some information text about the analyzis
var
_Start:Cardinal;
_WorkBitmap:TBitmap;
begin
  result:=false;
  ExifData:=TExif.create;
  Areas:=TRegionsList.Create;
  OriginalBitmap:=TBitmap.Create;
  OutputBitmap:=TBitmap.Create;
  _WorkBitmap:=TBitmap.Create;
  try
    if SameText(extractFileExt(_Filename),'.jpg') then begin
      OriginalBitmap.Free;
      if not Jpg2BMP(_Filename,OriginalBitmap) then begin
        AnalysisText:=format('Could not load file <%s>.',[_Filename]);
        exit;
      end;

      _WorkBitmap.free;
      ExifData.free;
      if not Jpg2BMP(_Filename,cImageSize,0,_WorkBitmap,ExifData) then begin
        AnalysisText:=format('Could not load file <%s>.',[_Filename]);
        exit;
      end;
    end else
    if SameText(extractFileExt(_Filename),'.bmp') then begin
      try
        OriginalBitmap.LoadFromFile(_Filename);
        //todo: implement resize and rotate
        _WorkBitmap.Assign(OriginalBitmap);
      except
        on e:exception do begin
          AnalysisText:=format('Could not load file <%s>. <%s>.',[_Filename,e.message]);
          exit;
        end;
      end;
    end
    else begin
      AnalysisText:='Only .jpg and .bmp files are supported.';
      exit;
    end;
    if not assigned(ExifData) then ExifData:=TExif.create;
    OutputBitmap.Assign(_WorkBitmap);


    FreeAndNil(Areas);
    _Start:=GetTickCount;
    FindFaceOnImage(_WorkBitmap,               // the input bitmap
                  ExifData,                    // _Exif data of the image (if available).
                  cFaceMinProbability,         // minimum Probability to look for. [0..100]
                  cMinWhitePixelsPercent,      // minimum number of white pixels [precent] inside a detected area.
                  cMaxWhitePixelsPercent,      // maximum number of white pixels [precent] inside a detected area.
                  cMinFaceRatio,               // minimum ratio width/height of a detected area.
                  cMaxFaceRatio,               // maximum ratio width/height of a detected area.
                  cSmallObjectsToRemove,       // remove small areas
                  true,
                  Areas,                       // returns a list with the detected areas.
                  Faces,                       // returns a list with the detected heads.
                  AnalysisText);               // returns some textual information about the analyis of the image.

    AnalysisText:=AnalysisText+#10+#13+format('Analyzing image took <%d> ms.',[GetTickCount-_Start]);
    DrawAreas(OutputBitmap,Faces,tdr_AreaBox);
    result:=true;
  finally
    if assigned(_WorkBitmap) then _WorkBitmap.Free;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: IsFaceOnImage
  Author:    sam
  Date:      11-Aug-2024
  Arguments: const _Filename:string; out Areas:TRegionsList; out Faces:TFaces; out AnalysisText:string
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function IsFaceOnImage( const _Filename:string;               // the image to be anaylzed (supports .bmp or .jpg files)
                        out Faces:TFaces):boolean;            // the detected faces
var
_WorkBitmap:TBitmap;
_Exif:TExif;
_Areas:TRegionsList;
_AnalysisText:string;
begin
  result:=false;
  _Areas:=nil;
  _Exif:=nil;
  _WorkBitmap:=nil;
  _AnalysisText:='';
  Faces:=TFaces.create;
  try
    if not SameText(extractFileExt(_Filename),'.jpg') then exit;
    if not Jpg2BMP(_Filename,650,0,_WorkBitmap,_Exif) then begin
      _Exif.free;
      _WorkBitmap.Free;
      exit;
    end;


    FreeAndNil(Faces);


    FindFaceOnImage(_WorkBitmap,               // the input bitmap
                  _Exif,                       // _Exif data of the image (if available).
                  cFaceMinProbability,         // minimum Probability to look for. [0..100]
                  cMinWhitePixelsPercent,      // minimum number of white pixels [precent] inside a detected area.
                  cMaxWhitePixelsPercent,      // maximum number of white pixels [precent] inside a detected area.
                  cMinFaceRatio,               // minimum ratio width/height of a detected area.
                  cMaxFaceRatio,               // maximum ratio width/height of a detected area.
                  cSmallObjectsToRemove,       // remove small areas
                  false,
                  _Areas,                      // returns a list with the detected areas.
                  Faces,                       // returns a list with the detected heads.
                  _AnalysisText);              // returns some textual information about the analyis of the image.



    result:=true;
  finally
    if assigned(_WorkBitmap) then FreeAndNil(_WorkBitmap);
    if assigned(_Exif) then FreeAndNil(_Exif);
    if assigned(_Areas) then FreeAndNil(_Areas);
  end;
end;

end.
