{-----------------------------------------------------------------------------
 Unit Name: uExifInfoReader
 Author:    sam
 Date:      01-Dez-2021
 Purpose: read  exif information
 History: this unit needs some help! Please help me to
          make this more robust. Or if the there is a better way to read
          Exif-Informaiton I would be very glad to hear.
-----------------------------------------------------------------------------}

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

unit uExifInfoReader;

interface

uses
  Classes,
  {$IFDEF FPC}
  SysUtils, Graphics
  {$ELSE}
  System.SysUtils,
  WinApi.Windows
  {$ENDIF}
  ;

const
  cMinFaceRatio= 0.35;              // min. width/height of the face area.
  cMinFaceHeight:integer=22;
  cMinFaceWidth:integer=14;

type
  TImageType=(tit_unknown,
            tit_indoor,      // artifical light
            tit_outdoor);    // natural light

  TEXIFValueType=(tvt_void,                //0
              tvt_unsigned_byte,       //1      1byte   value 0..255
              tvt_string,              //2      1byte
              tvt_unsigned_short,      //3      2bytes  value 0..65'535
              tvt_unsigned_long,       //4      4bytes  value 0..4'294'967'295
              tvt_unsigned_float,      //5      8bytes  frist 4 bytes are the numerator, the second 4 bytes, the denominator.
              tvt_signed_byte,         //6      1byte   value -128..127
              tvt_undefined,           //7      1byte
              tvt_signed_short,        //8      2bytes  value -32'768..32'767
              tvt_signed_long,         //9      4bytes  value -2'147'483'648..2'147'483'647
              tvt_signed_float,        //10     8bytes  frist 4 bytes are the numerator, the second 4 bytes, the denominator.
              tvt_single_float,        //11     4bytes
              tvt_double_float);       //12     8bytes


  TMarker = record
    Section  : Word;      //Section marker
    Len      : Word;      //Length Section
    Indefin  : Array [0..4] of Byte; //Indefiner - "Exif" 00, "JFIF" 00 and ets
    Pad      : Byte;      //0x00
  end;

  TTag = record
    TagID   : Word;            //Tag number
    TagType : Word;            //Type tag
    Count   : DWord;           //tag length
    OffSet  : DWord;           //Offset / Value
    SwapNeeded:boolean;
    function ToString:string;
    procedure Swap;
  end;

  TIFDHeader = record
    pad          : Byte; //00h
    ByteOrder    : Word; //II (4D4D)
    i42          : Word; //2A00
    IFD0offSet   : Cardinal; //0th offset IFD
    Interoperabil: Byte;
  end;

  TTraceMessageEvent=procedure(Sender:TObject;const _Level:Byte;const _msg:string;const _params:array of const) of object;

  TExif = class
    private
      FStream             : TStream;
      FImageDesc          : String;     //Picture description
      FMake               : String;     //Camera manufacturer
      FModel              : String;     //Camera model
      FOrientation        : System.Int16;//Image orientation - 1 normal
      FCopyright          : String;     //Copyright
      FValid              : Boolean;    //Has valid Exif header
      FDateTime           : String;     //Date and Time of Change
      FDateTimeOriginal   : String;     //Original Date and Time
      FDateTimeDigitized  : String;     //Camshot Date and Time
      FUserComments       : String;     //User Comments
      FIdfp               : System.UInt16;
      FObjectDistance     : double;
      FBrightness         : double;
      FLightSource        : System.UInt16;
      FISOSpeed           : System.Byte;
      FMeteringMode       : System.UInt16;
      FFlashInfo          : System.UInt16;
      FIdf                : TIFDHeader;
      FMarker             : TMarker;
      FImageWidth         : System.Int32;
      FImageHeight        : System.Int32;
      FFocalLength        : double;
      FExposureProg       : System.Byte;
      FFilename           : string;
      FOnTrace            : TTraceMessageEvent;
      FMinExpectedFaceHeight: integer;       // this value is not read from EXIF. But calculated using the EXIF-data.
      FMinExpectedFaceWidth: integer;        // this value is not read from EXIF. But calculated using the EXIF-data.
      function ReadStringValue(const _Tag:TTag;const _Offset: Cardinal;out Value:String): boolean;
      function ReadInt16Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.Int16):boolean;
      function ReadInt32Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.Int32):boolean;
      function ReadUInt16Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.UInt16):boolean;
      function ReadUInt32Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.UInt32):boolean;
      function ReadByteValue(const _Tag:TTag;const _Offset: Cardinal;out Value:System.Byte):boolean;
      function ReadUDoubleValue(const _Tag:TTag;const _Offset: Cardinal;out Value:double): boolean;
      function ReadDoubleValue(const _Tag:TTag;const _Offset: Cardinal;out Value:double): boolean;
      function ReadSingleValue(const _Tag:TTag;const _Offset: Cardinal;out Value:single): boolean;
      function ReadValue(const _Tag:TTag;const _Offset: Cardinal): Variant;
      procedure Init;
      function GetImageType: TImageType;
      procedure trace(const _Level:Byte;const _msg:string;const _params:array of const);
      procedure EstimateMinExpectedHeadSize;
    public
      constructor Create;
      destructor Destroy;override;
      function ReadFromFile(const _FileName: String):boolean;
      function FlashFired:boolean;
      function AsString:string;
      property ImageDesc: String read FImageDesc;
      property Make: String read FMake;
      property Model: String read FModel;
      property Orientation: System.Int16 read FOrientation;
      property Copyright: String read FCopyright;
      property Valid: Boolean read FValid;
      property DateTime: String read FDateTime;
      property DateTimeOriginal: String read FDateTimeOriginal;
      property DateTimeDigitized: String read FDateTimeDigitized;
      property UserComments: String read FUserComments;
      property ObjectDistance:double read FObjectDistance;
      property Brightness:double read FBrightness;
      property ImageType:TImageType read GetImageType;
      property LightSource:System.UInt16 read FLightSource;
      property MeteringMode: System.UInt16 read FMeteringMode;
      property FlashInfo:System.UInt16 read FFlashInfo;
      property ISOSpeed:System.Byte read FISOSpeed;
      property ImageWidth:System.Int32 read FImageWidth write FImageWidth;
      property ImageHeight:System.Int32 read FImageHeight write FImageHeight;
      property MinExpectedFaceHeight:integer read FMinExpectedFaceHeight;
      property MinExpectedFaceWidth:integer read FMinExpectedFaceWidth;
      property FocalLength:System.double read FFocalLength;
      property ExposureProg:System.Byte read FExposureProg;
      property OnTrace:TTraceMessageEvent read FOnTrace write FOnTrace;
  end;

implementation

uses
  TypInfo
  {$IFNDEF FPC}
  , VCL.Imaging.jpeg
  {$ENDIF}
  ;

const
  // EXIF Tags, from https://exiftool.org/TagNames/EXIF.html
  TAG_ImageWidth         = $0100;
  TAG_ImageHeight        = $0101;
  TAG_ImageDescription   = $010E;
  TAG_Make               = $010F;
  TAG_Model              = $0110;
  TAG_Orientation        = $0112;
    // 1 = Horizontal (normal), 2 = Mirror horizontal, 3 = Rotate 180
    // 4 = Mirror vertical, 5 = Mirror horizontal and rotate 270 CW
    // 6 = Rotate 90 CW, 7 = Mirror horizontal and rotate 90 CW
    // 8 = Rotate 270 CW
  TAG_DateTimeModified   = $0132;
  TAG_Copyright          = $8298;
  TAG_ExifIFD            = $8769;
  TAG_ExposureProgram    = $8822;
    // 0 = Not Defined, 1 = Manual, 2 = Program AE, 3 = Aperture-priority AE
    // 4 = Shutter speed priority AE, 5 = Creative (Slow speed), 6 = Action (High speed)
    // 7 = Portrait, 8 = Landscape, 9 = Bulb
  TAG_ISOSpeed           = $8827;
  TAG_DateTimeOriginal   = $9003;  // Date & time of original
  TAG_DateTimeDigitized  = $9004;  // Date & time of digital data generation
  TAG_BrightnessValue    = $9203;
  TAG_SubjectDistance    = $9206;
  TAG_MeteringMode       = $9207;
    // 0 = Unknown, 1 = Average, 2 = Center-weighted average, 3 = Spot, 4 = Multi-spot, 5 = Multi-segment, 6 = Partial, 255 = Other
  TAG_LightSource        = $9208;
    // 0 = Unknown, 1 = Daylight, 2 = Fluorescent, 3 = Tungsten, 4 = Flash, 9 = Fine weather, 10 = Cloudy, 11 = Shade,
    // 12 = Daylight Fluorescent, 13 = Day White Fluorescent, 14 = Cool White Fluorescent, 15 = White Fluorescent
    // 16 = Warm White Fluorescent, 17 = Standard Light A, 18 = Standard Light B, 19 = Standard Light C,
    // 20 = D55, 21 = D65, 22 = D75, 23 = D50, 24 = ISO Studio Tungsten, 255 = Other
  TAG_Flash              = $9209;   // see: https://exiftool.org/TagNames/EXIF.html#Flash
  TAG_FocalLength        = $920A;
  TAG_UserComment        = $9286;

function Swap16(Value: Word): Word;
begin
  result:=System.Swap(Value);
end;

function Swap32(_Value: dWord): dWord;
asm
  bswap eax;
end;

function EXIFValueTypeToStr(const _Value:TEXIFValueType):string;
begin
  result:=GetEnumName(TypeInfo(TEXIFValueType),integer(_Value));
  delete(result,1,4);
end;

{-----------------------------------------------------------------------------
  Procedure: ReadStringValue
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:String
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadStringValue(const _Tag:TTag;const _Offset: Cardinal;out Value:String): boolean;
var
_fp: Int64;
_tmp:AnsiString;
begin
  result:=false;
  Value:='';
  trace(5,'TExif.ReadStringValue: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  _fp:=FStream.Position;        // Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      SetLength(_tmp,_Tag.count-1);
      FStream.Read(_tmp[1],_Tag.count-1);
      Value:=string(_tmp);
      SetLength(_tmp,0);
      result:=true;
      trace(5,'TExif.ReadStringValue: Leave method with value <%s>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadStringValue: Could not read string-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;     // Restore file offset
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadInt16Value
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:System.Int16
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadInt16Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.Int16):boolean;
var
_fp: Int64;
begin
  result:=false;
  Value:=0;
  trace(5,'TExif.ReadInt16Value: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  if _Tag.Count<=4 then begin
    Value:=_Tag.OffSet;
    result:=true;
    trace(5,'TExif.ReadInt16Value: Leave method with value <%d>.',[Value]);
    exit;
  end;
  _fp:=FStream.Position;        // Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(Value, SizeOf(Value));
      if _Tag.SwapNeeded then Value := Swap16(Value);
      result:=true;
      trace(5,'TExif.ReadInt16Value: Leave method with value <%d>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadInt16Value: Could not read integer-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;     // Restore file offset
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadInt32Value
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:System.Int32
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadInt32Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.Int32):boolean;
var
_fp: Int64;
begin
  result:=false;
  Value:=0;
  trace(5,'TExif.ReadInt32Value: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  if _Tag.Count<=4 then begin
    Value:=_Tag.OffSet;
    result:=true;
    trace(5,'TExif.ReadInt32Value: Leave method with value <%d>.',[Value]);
    exit;
  end;
  _fp:=FStream.Position;        // Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(Value, SizeOf(Value));
      if _Tag.SwapNeeded then Value := Swap32(Value);
      result:=true;
      trace(5,'TExif.ReadInt32Value: Leave method with value <%d>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadInt32Value: Could not read integer-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;      // Restore file offset
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadUInt16Value
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:System.UInt16
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadUInt16Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.UInt16):boolean;
var
_fp: Int64;
begin
  result:=false;
  Value:=0;
  trace(5,'TExif.ReadUInt16Value: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  if _Tag.Count<=4 then begin
    Value:=_Tag.OffSet;
    result:=true;
    trace(5,'TExif.ReadUInt16Value: Leave method with value <%d>.',[Value]);
    exit;
  end;

  _fp:=FStream.Position;        // Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(Value, SizeOf(Value));
      if _Tag.SwapNeeded then Value := Swap16(Value);
      result:=true;
      trace(5,'TExif.ReadUInt16Value: Leave method with value <%d>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadUInt16Value: Could not read integer-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;             // Restore file offset
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadUInt32Value
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:System.UInt32
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadUInt32Value(const _Tag:TTag;const _Offset: Cardinal;out Value:System.UInt32):boolean;
var
_fp: Int64;
begin
  result:=false;
  Value:=0;
  trace(5,'TExif.ReadUInt32Value: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  if _Tag.Count<=2 then begin
    Value:=_Tag.OffSet;
    result:=true;
    trace(5,'TExif.ReadUInt32Value: Leave method with value <%d>.',[Value]);
    exit;
  end;
  _fp:=FStream.Position;        // Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(Value, SizeOf(Value)); // wp: was Value:=FBuffer.ReadInt16; --- why Int16?
      if _Tag.SwapNeeded then Value := Swap16(Value);      // wp: why 16 rather than 32?
      result:=true;
      trace(5,'TExif.ReadUInt32Value: Leave method with value <%d>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadUInt32Value: Could not read integer-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;    // Restore file offset
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadByteValue
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:System.Byte
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadByteValue(const _Tag:TTag;const _Offset: Cardinal;out Value:System.Byte):boolean;
var
_fp: Int64;
begin
  result:=false;
  Value:=0;
  trace(5,'TExif.ReadByteValue: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  _fp:=FStream.Position;        // Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(Value, SizeOf(Value));
      result:=true;
      trace(5,'TExif.ReadByteValue: Leave method with value <%d>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadByteValue: Could not read integer-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;     // Restore file offset
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadUDoubleValue
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:double
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadUDoubleValue(const _Tag:TTag;const _Offset: Cardinal;out Value:double): boolean;
var
_fp: Int64;
_Nominator:System.UInt32;
_Denominator:System.UInt32;
begin
  result:=false;
  trace(5,'TExif.ReadUDoubleValue: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  _fp:=FStream.Position;        //Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(_Nominator, SizeOf(_Nominator));
      FStream.Read(_Denominator, SizeOf(_Denominator));
      if _Tag.SwapNeeded then begin
        _Nominator:=Swap32(_Nominator);
        _Denominator:=Swap32(_Denominator);
      end;
      Value:= _Nominator/_Denominator;
      result:=true;
      trace(5,'TExif.ReadUDoubleValue: Leave method with value <%f>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadUDoubleValue: Could not read unsigned double-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;      // Restore file offset
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadDoubleValue
  Author:    sam
  Date:      27-Sep-2024
  Arguments: const _Tag:TTag;const _Offset: Cardinal;out Value:double
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadDoubleValue(const _Tag:TTag;const _Offset: Cardinal;out Value:double): boolean;
var
_fp: Int64;
_Nominator:System.Int32;
_Denominator:System.Int32;
begin
  result:=false;
  trace(5,'TExif.ReadDoubleValue: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  _fp:=FStream.Position;        //Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(_Nominator, SizeOf(_Nominator));
      FStream.Read(_Denominator, SizeOf(_Denominator));
      if _Tag.SwapNeeded then begin
        _Nominator:=Swap32(_Nominator);
        _Denominator:=Swap32(_Denominator);
      end;

      Value:= _Nominator/_Denominator;
      result:=true;
      trace(5,'TExif.ReadDoubleValue: Leave method with value <%f>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadDoubleValue: Could not read double-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;      // Restore file offset
  end;
end;


function TExif.ReadSingleValue(const _Tag:TTag;const _Offset: Cardinal;out Value:single): boolean;
var
_fp: Int64;
_Nominator:System.UInt16;
_Denominator:System.UInt16;
begin
  result:=false;
  trace(5,'TExif.ReadSingleValue: Enter method with tag <%s> and offset <%d>.',[_Tag.ToString,_Offset]);
  _fp:=FStream.Position;        //Save file offset
  try
    try
      FStream.Position:=_Tag.OffSet+_Offset;
      FStream.Read(_Nominator, SizeOf(_Nominator));
      FStream.Read(_Denominator, SizeOf(_Denominator));
      if _Tag.SwapNeeded then begin
        _Nominator:=Swap16(_Nominator);
        _Denominator:=Swap16(_Denominator);
      end;
      Value:= (_Nominator/_Denominator);
      result:=true;
      trace(5,'TExif.ReadSingleValue: Leave method with value <%f>.',[Value]);
    except
      on e:exception do trace(1,'Error in TExif.ReadSingleValue: Could not read single-data for tag.id <%x>. <%s>.',[_Tag.TagID,e.message]);
    end;
  finally
    FStream.Position:=_fp;      // Restore file offset
  end;
end;


function TExif.ReadValue(const _Tag:TTag;const _Offset: Cardinal): Variant;
var
_byteValue:Byte;
_StringValue:string;
_UInt16Value:System.UInt16;
_UInt32Value:System.UInt32;
_DoubleValue:double;
_Int16Value:System.Int16;
_Int32Value:System.Int32;
_SingleValue:Single;
begin
  case TEXIFValueType(_Tag.TagType) of
    tvt_unsigned_byte : if ReadByteValue(_Tag,_Offset,_byteValue)     then result:=_byteValue;
    tvt_string        : if ReadStringValue(_Tag,_Offset,_StringValue) then result:=_StringValue;
    tvt_unsigned_short: if ReadUInt16Value(_Tag,_Offset,_UInt16Value) then result:=_UInt16Value;     //3      2bytes  value 0..65'535
    tvt_unsigned_long : if ReadUInt32Value(_Tag,_Offset,_UInt32Value) then result:=_UInt32Value;     //4      4bytes  value 0..4'294'967'295
    tvt_unsigned_float: if ReadUDoubleValue(_Tag,_Offset,_DoubleValue) then result:=_DoubleValue;    //5      8bytes  frist 4 bytes are the numerator, the second 4 bytes, the denominator.
    tvt_signed_byte   : if ReadByteValue(_Tag,_Offset,_byteValue)     then result:=_byteValue;       //6      1byte   value -128..127
    tvt_undefined     :; //???           //7      1byte
    tvt_signed_short  : if ReadInt16Value(_Tag,_Offset,_Int16Value)      then result:=_Int16Value;   //3      2bytes  value 0..65'535        //8      2bytes  value -32'768..32'767
    tvt_signed_long   : if ReadInt32Value(_Tag,_Offset,_Int32Value)      then result:=_Int32Value;   //9      4bytes  value -2'147'483'648..2'147'483'647
    tvt_signed_float  : if ReadDoubleValue(_Tag,_Offset,_DoubleValue)    then result:=_DoubleValue;  //10     8bytes  frist 4 bytes are the numerator, the second 4 bytes, the denominator.
    tvt_single_float  : if ReadSingleValue(_Tag,_Offset,_SingleValue)    then result:=_SingleValue;  //11     4bytes
  end;
end;

function TExif.FlashFired: boolean;
begin
  result:=((FFlashInfo and 1)=1)
end;

function TExif.GetImageType: TImageType;
begin
  if FlashFired then result:=tit_indoor
                else result:=tit_outdoor;

end;

procedure TExif.Init;
begin
  FreeAndNil(FStream);
  FValid:=false;
  FIdfp:=0;
  FImageDesc:='';
  FMake:='';
  FModel:='';
  FOrientation:=1;
  FDateTime:='';
  FCopyright:='';
  FDateTimeOriginal:='';
  FDateTimeDigitized:='';
  FUserComments:='';
  FObjectDistance:=0;
  FISOSpeed:=0;
  FLightSource:=0;
  FBrightness:=0;
  FImageWidth:=0;
  FImageHeight:=0;
  FFocalLength:=0;
  FFilename:='';
  FMinExpectedFaceWidth:=cMinFaceWidth;
  FMinExpectedFaceHeight:=cMinFaceHeight;
end;

function TExif.AsString: string;
begin
  result:=FFilename+#13+#10;
  if not FValid then begin
    result:='no EXIF-Data available';
    exit;
  end;
  result:=result+'Description: '+FImageDesc+#13+#10;
  result:=result+'Make: '+FMake+#13+#10;
  result:=result+'Model: '+FModel+#13+#10;
  result:=result+'Orientation: '+FOrientation.ToString+#13+#10;
  result:=result+'DateTime: '+FDateTime+#13+#10;
  result:=result+'Object-Distance: '+FObjectDistance.ToString+#13+#10;
  result:=result+'ExposureProg: '+FExposureProg.ToString+#13+#10;
  result:=result+'ISO-Speed: '+FISOSpeed.ToString+#13+#10;
  result:=result+'Light-Source: '+FLightSource.ToString+#13+#10;
  result:=result+'Brightness: '+FBrightness.ToString+#13+#10;
  if FlashFired then result:=result+'Flash used: yes'+#13+#10
                else result:=result+'Flash used: no'+#13+#10;
  result:=result+'Image size [pixel]: w'+inttostr(FImageWidth)+' x h'+inttostr(FImageHeight)+#13+#10;
  result:=result+'Min. expected face size [pixel]: w'+ FMinExpectedFaceWidth.ToString+' x h'+FMinExpectedFaceHeight.ToString+#13+#10;
  result:=result+'FocalLength: '+FFocalLength.ToString+#13+#10;
end;

constructor TExif.Create;
begin
  Init;
end;

destructor TExif.destroy;
begin
  FStream.Free;
  inherited;
end;

{-----------------------------------------------------------------------------
  Procedure: ReadFromFile
  Author:    sam
  Date:      17-Aug-2024
  Arguments: const _FileName: String
  Result:    boolean
  Description:
-----------------------------------------------------------------------------}
function TExif.ReadFromFile(const _FileName: String):boolean;
var
i: Integer;
_off0: Cardinal; // Null Exif Offset
_tag: TTag;
_StartofImage: Word; //2 bytes SOI marker. FF D8 (Start Of Image)
_JFIFfound:boolean;
begin
  result:=false;
  _JFIFfound:=false;
  Init;
  if not FileExists(_FileName) then begin
    trace(3,'Warning in TExif.ReadFromFile: Could not find file <%s>',[_FileName]);
    exit;
  end;
  FStream := TFileStream.Create(_FileName, fmOpenRead);
  try
    try
      FStream.Position:=0;
      FStream.Read(_StartOfImage, Sizeof(_StartOfImage));
      if _StartofImage<>$D8FF then exit;     // Is this Jpeg
      FStream.Read(FMarker, 9);

      if FMarker.Section=$E0FF then begin // JFIF Marker Found
        FStream.Seek(9,soCurrent);  //TODO: I do not understand this code.     //Skip JFIF Header
        FStream.Read(FMarker, 9);
        _JFIFfound:=true;
      end;

      if FMarker.Section<>$E1FF then begin
        result:=true;
        exit;
      end;

      FValid:=True;                // If we found Exif Section. j.Indefin='Exif'.

      _off0:=FStream.Position+1;
      FStream.Read(FIdf,11);       // Read IDF Header
       i:=0;
       repeat
        inc(i);
        FStream.Read(_Tag,12);
        if FIdf.ByteOrder=77 then _Tag.Swap;
        trace(3,'TExif.ReadFromFile: Read tag <%s>',[_Tag.ToString]);
        case _Tag.TagID of
          TAG_ImageDescription : FImageDesc:=ReadValue(_Tag,_off0);  //0E01 ImageDescription
          TAG_Make             : FMake:=ReadValue(_Tag,_off0);       //0F01 Make
          TAG_Model            : FModel:=ReadValue(_Tag,_off0);      //1001 Model
          TAG_ExifIFD          : FIdfp:=_Tag.OffSet;
          TAG_DateTimeModified : FDateTime:=ReadValue(_Tag,_off0);   //3201 DateTime
          TAG_Copyright        : FCopyright:=ReadValue(_Tag,_off0);  //9882 CopyRight
          TAG_Orientation      : FOrientation:=_tag.OffSet;
        end;
      until (i>30);
      if FIdfp>0 then begin
        FStream.Position:=FIdfp+12+2;  //12 - Size header before Exif, 2 - size Exif IFD Number
        if _JFIFfound then begin
          FStream.Position:=FStream.Position+11+7;
          _off0:=0;
        end;

        i:=0;
        repeat
          inc(i);
          FStream.Read(_Tag,12);
          if FIdf.ByteOrder=77 then _Tag.Swap;
          trace(3,'TExif.ReadFromFile: Read tag <%s>',[_Tag.ToString]);
            case _tag.TagID of
              TAG_ExposureProgram  :FExposureProg      := ReadValue(_tag,_off0); //_tag.OffSet;
              TAG_DateTimeOriginal : FDateTimeOriginal  := ReadValue(_tag,_off0);       //0390 FDateTimeOriginal
              TAG_SubjectDistance  : FObjectDistance    := ReadValue(_tag,_off0);
              TAG_LightSource      : FLightSource       := ReadValue(_tag,_off0); //_tag.OffSet;
              TAG_DateTimeDigitized: FDateTimeDigitized := ReadValue(_tag,_off0);       //0490 DateTimeDigitized
              TAG_UserComment      : FUserComments      := ReadValue(_tag,_off0);       //8692 UserComments
              TAG_ISOSpeed         : FISOSpeed          := ReadValue(_tag,_off0); //_tag.OffSet;
              TAG_BrightnessValue  : FBrightness        := ReadValue(_tag,_off0);
              TAG_MeteringMode     : FMeteringMode      := ReadValue(_tag,_off0);
              TAG_Flash            : FFlashInfo         := ReadValue(_tag,_off0); //_tag.OffSet;
              TAG_ImageWidth       : FImageWidth        := ReadValue(_tag,_off0); //_tag.OffSet;
              TAG_ImageHeight      : FImageHeight       := ReadValue(_tag,_off0); //_tag.OffSet;
              TAG_FocalLength      : FFocalLength       := ReadValue(_tag,_off0);
            end;
            if (FImageWidth<1000) and (FImageWidth>400) then begin
              break;
            end;

        until (i>40);
      end;
      FFilename:=_FileName;
      EstimateMinExpectedHeadSize;
      result:=true;
    except
      on e:exception do trace(1,'Error in TExif.ReadFromFile: Could not read from file <%s>. <%s>.',[_FileName,e.message]);
    end;
  finally
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: EstimateMinExpectedHeadSize
  Author:    sam
  Date:      17-Aug-2024
  Arguments: None
  Result:    None
  Description: try to calculate the minimum expected height and witdh of
               a face according to the available exif-data.
-----------------------------------------------------------------------------}
procedure TExif.EstimateMinExpectedHeadSize;
const
cAverageFaceHeight = 0.24; // Durchschnittliche Gesichtshhe in Metern (24 cm)
cSensorHeight35mm = 0.024; // Sensorhhe fr 35mm-Film in Metern (24mm)
begin
  FMinExpectedFaceWidth:=cMinFaceWidth;
  FMinExpectedFaceHeight:=cMinFaceHeight;
  if not FValid then exit;
  if (FFocalLength>0.1) and (FFocalLength<30) then begin
    FMinExpectedFaceHeight := round((cAverageFaceHeight * FFocalLength/1000 * FImageHeight) / cSensorHeight35mm /2);
    FMinExpectedFaceWidth  := round(FMinExpectedFaceHeight * cMinFaceRatio);
  end;
  if FMinExpectedFaceWidth<cMinFaceWidth   then FMinExpectedFaceWidth:=cMinFaceWidth;
  if FMinExpectedFaceHeight<cMinFaceHeight then FMinExpectedFaceHeight:=cMinFaceHeight;
end;


{ TTag }

procedure TTag.Swap;
begin
  TagID  :=Swap16(TagID);
  TagType:=Swap16(TagType);
  Count  :=Swap32(Count);
  if Count<=4 then begin
    case TagType of
      1,3: Offset  := Swap16(Offset);
      else Offset  := Swap32(Offset);
    end;
  end;
  SwapNeeded:=true;
end;

function TTag.ToString: string;
begin
  result:=format('ID: <%x>, Type: <%s>, Count <%d>, Offset <%d>.' ,[TagID,EXIFValueTypeToStr(TEXIFValueType(TagType)),Count,OffSet]);
end;

procedure TExif.trace(const _Level: Byte; const _msg: string;const _params: array of const);
begin
  if assigned(FOnTrace) then FOnTrace(Self,_Level,_msg,_params)
{$ifdef debug}
                        else outputdebugstring(PChar(Format(_msg,_params)));
{$endif}
end;




end.
