{-----------------------------------------------------------------------------
 Unit Name: uImageManipulation
 Author:    sam
 Date:      20-Nov-2017
 Purpose:   this unit contains some image manipulation routines.
 History:
-----------------------------------------------------------------------------}
unit uImageManipulation;

{$IFDEF FPC}
 {$MODE Delphi}
{$ENDIF}

interface

uses
 {$IFDEF FPC}
  Graphics, ExtCtrls, Types, Generics.Collections,
 {$ELSE}
  WinApi.Windows,
  Vcl.Graphics,
  VCL.ExtCtrls,
  System.Types,
  System.Generics.Collections,
 {$ENDIF}
  uImageDefs,
  uExifInfoReader,
  uAreaDetection,
  uDetectFace;


function AutoBlackWhiteImage(var Bitmap: TBitmap;const _Exif:TExif;out ObjectArea:TRect):boolean;
function Crop(var Bitmap: TBitmap;const _SourceRect:TRect):boolean;
function DetermBackgroundColor(const _Bitmap:TBitmap;out red_min,red_max,green_min,green_max,blue_min,blue_max:byte):boolean;
procedure RotateBitmap(_SrcImage,_TrgImage:TBitmap;const _ImageRotate:TImageRotate);
procedure SmoothResize(Src, Dst: TBitmap;_NewWidth,_NewHeight:integer);
procedure BlackWhiteImage(var Bitmap:TBitmap; const _threshold: byte;out ObjectArea:TRect);
function RemoveBackgroundColor(var Bitmap:TBitmap; const _threshold: byte):boolean;
procedure ManipulateImage(var Bitmap:TBitmap;const _ImageOperation:TImageOperation);
function Jpg2BMP(const _JpgFilename: String;out BMP:TBitmap): boolean;overload;
function  Jpg2BMP(const _JpgFilename: String;_NewWidth,_NewHeight:integer;out BMP:TBitmap;out ExifData:TExif): boolean;overload;
procedure DetermBackgroundValue(const _Bitmap:TBitmap; out backgroundValue: byte);overload;
procedure DetermBackgroundValue(const _Bitmap:TBitmap;const _Exif:TExif; out backgroundValue: byte);overload;
procedure DetermForegroundValue(const _Bitmap:TBitmap; out foregroundValue: byte);
function RemoveSmallObjects(var Bitmap:TBitmap;const _WorkArea:TRect;const _ObjectHeight,_ObjectWidth:integer;const _Treshold:integer;out ObjectArea:TRect):boolean;
procedure DrawRect(var Bitmap:TBitmap;const _Rect:TRect;const _Color:TColor;_BorderWidth:byte);
procedure DrawCross(var Bitmap:TBitmap;const _CenterPoint:TPoint;const _Color:TColor);
procedure DrawAreas(var B:TBitmap; _AreaList: TRegionsList;const _DrawType:TDrawType);overload;
procedure DrawAreas(var B:TBitmap; _Faces:TFaces;const _DrawType:TDrawType);overload;
procedure DrawRegion(var B:TBitmap;const _Region:TRegion;const _DrawType:TDrawType);
function LoadImage(const _JPGFilename:string;out Image:TImage):boolean;
function GetBrightness(var Bitmap:TBitmap;const _Rect:TRect;out Brightness:double;out DominantColor:byte):Boolean;
function ConvertToGrayscale(const _Source: TBitmap;out Target:TBitmap): boolean;
function ResizeBitmap(const _Source: TBitmap; const _Width, _Height: Integer;out Target:TBitmap): boolean;


implementation

uses
  SysUtils,
  Math,
  Dialogs,
  {$IFDEF FPC}
  Classes, Generics.Defaults, FPReadJPEG, LCLType, LCLIntf
 {$ELSE}
  System.Classes,
  System.Generics.Defaults,
  JPeg
 {$ENDIF}
  ;


const
  PixelCountMax = 32768;

type
  pRGBArray = ^TRGBArray;
  TRGBArray = array[0..PixelCountMax-1] of TRGBTriple;

{-----------------------------------------------------------------------------
  Procedure: LoadImage
  Author:    sam
  Date:      02-Dez-2021
  Arguments: const _Filename:string;var Image:TImage
  Result:    boolean
  Description: load the file <_filename> into image <_image>.
  This method reads the exif-information and then rotates the image if needed.
-----------------------------------------------------------------------------}
function LoadImage(const _JPGFilename:string;out Image:TImage):boolean;
var
_Rotate:TImageRotate;
_SrcImage:TBitmap;
_TrgImage:TBitmap;
_ExifData:TExif;
begin
  result:=false;
  _Rotate:=tr0;
  Image:=nil;
  if not Jpg2BMP(_JPGFilename,0,0,_SrcImage,_ExifData) then exit;
  if assigned(_ExifData) then _Rotate:=ExifOrientationToRotate(_ExifData.Orientation);
  _TrgImage:=TBitmap.Create;
  _TrgImage.PixelFormat:=pf24bit;
  try
    RotateBitmap(_SrcImage,_TrgImage,_Rotate);
    Image.Picture.Assign(_TrgImage);
    result:=true;
  finally
    _SrcImage.Free;
    _TrgImage.Free;
    if assigned(_ExifData) then _ExifData.Free;
  end;
end;

function RemoveBackgroundColor(var Bitmap:TBitmap; const _threshold: byte):boolean;
var
y,x: integer;
_Pixels:TPixelMap;
   function GetPxelBrightness(const _x,_y:integer):Byte;
  begin
    result:=0;
    if (_x<0) or (_x>=Bitmap.Width) then exit;
    if (_y<0) or (_y>=Bitmap.Height) then exit;
    result:=round(((_Pixels[_y,_x].rgbtRed+_Pixels[_y,_x].rgbtGreen+_Pixels[_y,_x].rgbtBlue) div 3));
  end;

  procedure MakeBlackAndWhite(const _x,_y:integer);
  begin
    if GetPxelBrightness(_x,_y)>=_threshold then exit;
    _Pixels[_y,_x].rgbtred:=0;
    _Pixels[_y,_x].rgbtgreen:=0;
    _Pixels[_y,_x].rgbtblue:=0;
  end;

begin
  result:=false;
  if _threshold=0 then exit;
  if not Bitmap2PixelMap(Bitmap,false,_Pixels) then exit;
  try
    for y:= 0 to Bitmap.height-1 do begin
      for x:= 0 to Bitmap.Width-1 do MakeBlackAndWhite(x,y);
    end;
    result:=true;
  finally
    PixelMap2Bitmap(_Pixels,Bitmap);
    SetLength(_Pixels,0);
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: GetBrightness
  Author:    sam
  Date:      29-Dez-2021
  Arguments: var Bitmap:TBitmap; const _Rect:TRect;out Brightness:double
  Result:    Boolean
  Description: returns brightness in % of the area defined by <_Rect>.
               100% means white. 0% means black.
-----------------------------------------------------------------------------}
function GetBrightness(var Bitmap:TBitmap;const _Rect:TRect;out Brightness:double;out DominantColor:byte):Boolean;
var
_Pixels:TPixelMap;
x,y:integer;
_SumRed:Cardinal;
_SumGreen:Cardinal;
_SumBlue:Cardinal;

function GetPixelBrightness(const _x,_y:integer):Byte;
begin
  result:=0;
  if (_x<0) or (_x>=Bitmap.Width) then exit;
  if (_y<0) or (_y>=Bitmap.Height) then exit;
  result:=round(((_Pixels[_y,_x].rgbtRed*3)+(4*_Pixels[_y,_x].rgbtGreen)+_Pixels[_y,_x].rgbtBlue) shr 3);
  _SumRed:=_SumRed+_Pixels[_y,_x].rgbtRed;
  _SumGreen:=_SumGreen+_Pixels[_y,_x].rgbtGreen;
  _SumBlue:=_SumBlue+_Pixels[_y,_x].rgbtBlue;
end;

begin
  result:=false;
  if (_Rect.Width=0) or
     (_Rect.Height=0) then exit;
  _SumRed:=0;
  _SumGreen:=0;
  _SumBlue:=0;
  DominantColor:=0;
  Brightness:=0;
  if not Bitmap2PixelMap(Bitmap,false,_Pixels) then exit;
  for y:= _Rect.top to _Rect.Bottom do begin
    for x:=_Rect.Left to _Rect.Right do Brightness:=Brightness+GetPixelBrightness(x,y);
  end;
  Brightness:=Brightness/(_Rect.Width*_Rect.Height*255);
  if (_SumRed>_SumGreen) and
     (_SumRed>_SumBlue) then  DominantColor:=1 else
  if (_SumGreen>_SumRed) and
     (_SumGreen>_SumBlue) then  DominantColor:=2 else
  if (_SumBlue>_SumRed) and
     (_SumBlue>_SumGreen) then  DominantColor:=3;
  result:=true;
  SetLength(_Pixels,0);
end;

{-----------------------------------------------------------------------------
  Procedure: UpperPartIsBright
  Author:    sam
  Date:      29-Dez-2021
  Arguments: var Bitmap:TBitmap;const _MinBrightness:byte
  Result:    boolean
  Description: Returns true if the upper half of the image is brighter than
               the lower half of the image.
               <_MinBrightness> defines the minimum brightness [%] of the upper half.
-----------------------------------------------------------------------------}
function UpperPartIsBright(var Bitmap:TBitmap;const _MinBrightness:double):boolean;
var
_UpperPart:TRect;
_BrightnessUpperPart:double;
_LowerPart:TRect;
_BrightnessLowerPart:double;
_DominantColor:byte;
begin
  result:=false;
// calculate brightness of the upper half image.
  _UpperPart.Left:=0;
  _UpperPart.Right:=Bitmap.Width;
  _UpperPart.Top:=0;
  _UpperPart.Bottom:=Bitmap.Width div 2;
  if not GetBrightness(Bitmap,_UpperPart,_BrightnessUpperPart,_DominantColor) then exit;
  if _BrightnessUpperPart<_MinBrightness then exit;
  if _DominantColor=2 then exit;   // if dominant color is green, then  it might not a sunset

// calculate brightness of the lower half image.
  _LowerPart.Left:=0;
  _LowerPart.Right:=Bitmap.Width;
  _LowerPart.Top:=(Bitmap.Width div 2)+1;
  _LowerPart.Bottom:=Bitmap.height;
  if not GetBrightness(Bitmap,_LowerPart,_BrightnessLowerPart,_DominantColor) then exit;

  result:=(_BrightnessUpperPart>_BrightnessLowerPart);
end;

{-----------------------------------------------------------------------------
  Procedure: DrawRect
  Author:    sam
  Date:      27-Jul-2022
  Arguments: var Bitmap:TBitmap;const _Rect:TRect;const _Color:TColor;_BorderWidth:byte
  Result:    None
  Decription: draw the rectangle <_Rect> onto bitmap <Bitmap>.
-----------------------------------------------------------------------------}
procedure  DrawRect(var Bitmap:TBitmap;const _Rect:TRect;const _Color:TColor;_BorderWidth:byte);
var
y,x: integer;
_Pixels:TPixelMap;
_RGB:TColor;
_R,_G,_B:byte;
_bw:byte;

  procedure ManipulatePixel(const _x,_y:integer);
  begin
    if _x<0 then exit;
    if _x>Bitmap.Width-1 then exit;
    if _y<0 then exit;
    if _y>Bitmap.Height-1 then exit;
    if _Pixels[_y,_x].rgbtred<>_R   then _Pixels[_y,_x].rgbtred:=_R;
    if _Pixels[_y,_x].rgbtgreen<>_G then _Pixels[_y,_x].rgbtgreen:=_G;
    if _Pixels[_y,_x].rgbtblue<>_B  then _Pixels[_y,_x].rgbtblue:=_B;
  end;

begin
  _RGB:=ColorToRGB(_Color);
  _R:=(_RGB and $FF);
  _G:=(_RGB and $FF00) shr 8;
  _B:=(_RGB and $FF0000) shr 16;
  if not Bitmap2PixelMap(Bitmap,false,_Pixels) then exit;
  try
    for x:= _Rect.Left to _Rect.Right do begin
      for _bw := 0 to _BorderWidth-1 do begin
        ManipulatePixel(x,_Rect.top-_bw);
        ManipulatePixel(x,_Rect.Bottom+_bw);
      end;
    end;
    for y:= _Rect.Top to _Rect.Bottom do begin
      for _bw := 0 to _BorderWidth-1 do begin
        ManipulatePixel(_Rect.Right+_bw,y);
        ManipulatePixel(_Rect.Left-_bw,y);
      end;
    end;
  finally
    PixelMap2Bitmap(_Pixels,Bitmap);
    SetLength(_Pixels,0);
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: DrawCross
  Author:    sam
  Date:      27-Jul-2022
  Arguments: var Bitmap:TBitmap;const _CenterPoint:TPoint;const _Color:TColor
  Result:    None
  Description: draw a cross onto bitmap <Bitmap> at center point <_CenterPoint>
               with color <_Color>.
-----------------------------------------------------------------------------}
procedure  DrawCross(var Bitmap:TBitmap;const _CenterPoint:TPoint;const _Color:TColor);
var
y,x: integer;
_Pixels:TPixelMap;
_RGB:TColor;
_R,_G,_B:byte;

  procedure ManipulatePixel(const _x,_y:integer);
  begin
    if (_x<0) or (_x>=Bitmap.Width) then exit;
    if (_y<0) or (_y>=Bitmap.Height) then exit;
    if _Pixels[_y,_x].rgbtred<>_R then _Pixels[_y,_x].rgbtred:=_R;
    if _Pixels[_y,_x].rgbtgreen<>_G then _Pixels[_y,_x].rgbtgreen:=_G;
    if _Pixels[_y,_x].rgbtblue<>_B then _Pixels[_y,_x].rgbtblue:=_B;
  end;

begin
  _RGB:=ColorToRGB(_Color);
  _R:=(_RGB and $FF);
  _G:=(_RGB and $FF00) shr 8;
  _B:=(_RGB and $FF0000) shr 16;
  if not Bitmap2PixelMap(Bitmap,false,_Pixels) then exit;
  try
    for x:= _CenterPoint.x-20 to _CenterPoint.x+20 do ManipulatePixel(x,_CenterPoint.y);
    for y:= _CenterPoint.y-20 to _CenterPoint.y+20 do ManipulatePixel(_CenterPoint.x,y);
  finally
    PixelMap2Bitmap(_Pixels,Bitmap);
    SetLength(_Pixels,0);
  end;
end;


{-----------------------------------------------------------------------------
  Procedure: DrawAreas
  Author:    sam
  Date:      27-Jul-2022
  Arguments: var _B:TBitmap; _AreaList: TRegionsList;const _DrawType:TDrawType
  Result:    None
  Description: draw a rect for each area in list <_AreaList>.
-----------------------------------------------------------------------------}
procedure DrawAreas(var B:TBitmap; _AreaList: TRegionsList;const _DrawType:TDrawType);
var
i:integer;
_ObjectArea: TRegion;
begin
  if not assigned(_AreaList) then exit;
  for i:=0  to  _AreaList.Count-1 do begin
    _ObjectArea:=_AreaList.Items[i];
    DrawRegion(B,_ObjectArea,_DrawType);
  end;
end;

procedure DrawAreas(var B:TBitmap; _Faces:TFaces;const _DrawType:TDrawType);
var
i:integer;
_ObjectArea: TRegion;
begin
  if not assigned(_Faces) then exit;
  for i:=0  to _Faces.Count-1 do begin
    _ObjectArea:=_Faces.Items[i].head;
    DrawRegion(B,_ObjectArea,_DrawType);
  end;
end;

procedure DrawRegion(var B:TBitmap;const _Region:TRegion;const _DrawType:TDrawType);
var
_Color:TColor;
_Text:string;
begin
  _text:='';
  case _DrawType of
    tdr_CenterPoint: begin
       DrawRect(B,  _Region.CenterRect, clYellow, 2);
       DrawRect(B,  _Region.Rect , clBlue, 1);
       DrawCross(B, _Region.FocusPoint, clBlue);
    end;
    tdr_AreaBox: begin
       if _Region.Probability=0 then _Color:=clYellow
                                else _Color:=trunc(clYellow*_Region.Probability);

       B.Canvas.Pen.Color:=_Color;
       B.Canvas.Pen.Width:=2;
       B.Canvas.Brush.Style:=bsClear;
       B.Canvas.Rectangle(_Region.Rect);
    end;
    tdr_CenterRect: begin
       B.Canvas.Pen.Color:=clBlue;
       B.Canvas.Pen.Width:=4;
       B.Canvas.Brush.Style:=bsClear;
       B.Canvas.Rectangle(_Region.CenterRect);
    end;
  end;
  if _Region.Probability>0 then _text:=format('%d (%d)',[_Region.No,_Region.Probability])
                           else _text:=format('%d',[_Region.No]);
  B.Canvas.Font.Size:=14;
  B.Canvas.Font.Color:=clYellow;
  B.Canvas.Font.Style:=[TFontStyle.fsBold];
  B.Canvas.TextOut(_Region.Rect.Left,_Region.Rect.Top-28,_text);
end;

{-----------------------------------------------------------------------------
  Procedure: RemoveSmallObjects
  Author:    sam
  Date:      31-Mai-2021
  Arguments: var Bitmap:TBitmap;const _ObjectHeight,_ObjectWidth:integer
  Result:    None
  Description: Calculate the average around a pixel. If the pixel
               is darker than then the average, then set the pixel to the average.

               Parameter <Bitmap> defines the bitmap to work on.
               Parameter <_WorkArea> defines the area inside the bitmap to work on.
               Parameter <_ObjectHeight,_ObjectWidth> define the size of the
               objects to be removed.
-----------------------------------------------------------------------------}
function RemoveSmallObjects(var Bitmap:TBitmap;
                            const _WorkArea:TRect;
                            const _ObjectHeight,_ObjectWidth:integer;
                            const _Treshold:integer;
                            out ObjectArea:TRect):boolean;
var
y,x: integer;
_Pixels:TPixelMap;
_w,_h:integer;
_RelevantPixelThreshold:integer;

  procedure ManipulatePixel(const _x,_y:integer;const _white:boolean);
  var
  x,y:integer;
  _NoOfRelevantPixels:integer;
  begin
    _NoOfRelevantPixels:=0;
    for x := _x-_w to _x+_w do begin
      for y := _y-_h to _y+_h do begin
        if (x>0) and
           (x<Bitmap.Width) and
           (y>0) and
           (y<Bitmap.Height) then begin
             if _white then begin
               if (_Pixels[y,x].rgbtred=255) then begin
                 inc(_NoOfRelevantPixels);
                 if _NoOfRelevantPixels>_RelevantPixelThreshold then begin
                   if _x<ObjectArea.Left   then ObjectArea.Left :=_x;
                   if _x>ObjectArea.Right  then ObjectArea.Right:=_x;
                   if _y<ObjectArea.Top    then ObjectArea.Top  :=_y;
                   if _y>ObjectArea.Bottom then ObjectArea.Bottom:=_y;
                   exit;
                 end;
               end;
             end
             else begin
               if (_Pixels[y,x].rgbtred=0) then begin
                 inc(_NoOfRelevantPixels);
                 if _NoOfRelevantPixels>15 then begin
                   if _x<ObjectArea.Left   then ObjectArea.Left :=_x;
                   if _x>ObjectArea.Right  then ObjectArea.Right:=_x;
                   if _y<ObjectArea.Top    then ObjectArea.Top  :=_y;
                   if _y>ObjectArea.Bottom then ObjectArea.Bottom:=_y;
                   exit;
                 end;
               end;
             end;
        end
      end;
    end;
    if _white then begin
      if _Pixels[_y,_x].rgbtred>0   then _Pixels[_y,_x].rgbtred:=0;
      if _Pixels[_y,_x].rgbtgreen>0 then _Pixels[_y,_x].rgbtgreen:=0;
      if _Pixels[_y,_x].rgbtblue>0  then _Pixels[_y,_x].rgbtblue:=0;
    end
    else begin
      if _y>0 then begin
        if _Pixels[_y-1,_x].rgbtred<255   then _Pixels[_y-1,_x].rgbtred:=0;
        if _Pixels[_y-1,_x].rgbtgreen<255 then _Pixels[_y-1,_x].rgbtgreen:=0;
        if _Pixels[_y-1,_x].rgbtblue<255  then _Pixels[_y-1,_x].rgbtblue:=0;
      end;

      if _y<Bitmap.Height-1 then begin
        if _Pixels[_y+1,_x].rgbtred<255   then _Pixels[_y+1,_x].rgbtred:=0;
        if _Pixels[_y+1,_x].rgbtgreen<255 then _Pixels[_y+1,_x].rgbtgreen:=0;
        if _Pixels[_y+1,_x].rgbtblue<255  then _Pixels[_y+1,_x].rgbtblue:=0;
      end;


      if _Pixels[_y,_x].rgbtred<255   then _Pixels[_y,_x].rgbtred:=0;
      if _Pixels[_y,_x].rgbtgreen<255 then _Pixels[_y,_x].rgbtgreen:=0;
      if _Pixels[_y,_x].rgbtblue<255  then _Pixels[_y,_x].rgbtblue:=0;

      if _x>0 then begin
        if _Pixels[_y,_x-1].rgbtred<255   then _Pixels[_y,_x-1].rgbtred:=0;
        if _Pixels[_y,_x-1].rgbtgreen<255 then _Pixels[_y,_x-1].rgbtgreen:=0;
        if _Pixels[_y,_x-1].rgbtblue<255  then _Pixels[_y,_x-1].rgbtblue:=0;
      end;

      if _x<Bitmap.Width-1 then begin
        if _Pixels[_y,_x+1].rgbtred<255   then _Pixels[_y,_x+1].rgbtred:=0;
        if _Pixels[_y,_x+1].rgbtgreen<255 then _Pixels[_y,_x+1].rgbtgreen:=0;
        if _Pixels[_y,_x+1].rgbtblue<255  then _Pixels[_y,_x+1].rgbtblue:=0;
      end;

    end;
  end;

begin
  result:=false;
  ObjectArea.Left:=0;
  ObjectArea.Top:=0;
  ObjectArea.Right:=0;
  ObjectArea.Bottom:=0;
  if (_ObjectHeight=0) and (_ObjectWidth=0) then begin
    result:=true;
    exit;
  end;
  if not Bitmap2PixelMap(Bitmap,false,_Pixels) then exit;
  ObjectArea.Left:=65000;
  ObjectArea.Top:=65000;
  ObjectArea.Right:=-1;
  ObjectArea.Bottom:=-1;
  try
    _h:= _ObjectHeight div 2;
    _w:= _ObjectWidth div 2;
    if _Treshold=0 then _RelevantPixelThreshold:=trunc(_ObjectHeight*_ObjectWidth*0.4)
                   else _RelevantPixelThreshold:=_Treshold;
    for y:= _WorkArea.top+1 to _WorkArea.Bottom-1 do begin
      for x:=_WorkArea.Left+1 to _WorkArea.right-1 do ManipulatePixel(x,y,true);
    end;

    for y:= _WorkArea.top+1 to _WorkArea.Bottom-1 do begin
      for x:=_WorkArea.Left+1 to _WorkArea.right-1 do ManipulatePixel(x,y,false);
    end;


    ObjectArea.Left:=ObjectArea.Left-1;
    ObjectArea.Top:=ObjectArea.Top-1;
    ObjectArea.Right:= ObjectArea.Right+1;
    ObjectArea.Bottom:=ObjectArea.Bottom+1;
    result:=true;
  finally
    PixelMap2Bitmap(_Pixels,Bitmap);
    SetLength(_Pixels,0);
  end;
end;

function Jpg2BMP(const _JpgFilename: String;out BMP:TBitmap): boolean;
var
_Jpg:TJpegImage;
begin
  result:=false;
  BMP:=TBitmap.create;
  if _JpgFilename='' then exit;
  if not fileexists(_JpgFilename) then exit;
  _Jpg:=TJpegImage.Create;
  try
    _Jpg.Performance:=jpBestSpeed;
    {$IFNDEF FPC}
    _Jpg.PixelFormat:=jf24Bit;
    {$ENDIF}
    _Jpg.LoadFromFile(_JpgFilename);
    if (_Jpg.Width=0) or
       (_Jpg.Height=0) then exit;

    if (_Jpg.Width>4000)  or (_Jpg.Height>4000) then _Jpg.Scale := jsEighth  else
    if (_Jpg.Width>2000)  or (_Jpg.Height>2000) then _Jpg.Scale := jsQuarter else
    if (_Jpg.Width>1000)  or (_Jpg.Height>1000) then _Jpg.Scale := jsHalf;
    BMP.Assign(_Jpg);
    result:=true;
  finally
    _Jpg.Free;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: Jpg2BMP
  Author:    sam
  Date:      31-Mai-2021
  Arguments: const _JpgFilename: String;_NewWidth,_NewHeight:integer;out BMP:TBitmap
  Result:    boolean
  Description: load a the file <_JpgFilename> into the out-parameter <BMP>.
               Parameters <_NewWidth,_NewHeight> can be used to define the size.
               The method will also try to rotate the image if exif-information
               is available.
-----------------------------------------------------------------------------}
function Jpg2BMP(const _JpgFilename: String;_NewWidth,_NewHeight:integer;out BMP:TBitmap;out ExifData:TExif): boolean;
var
_Bitmap: TBitmap;
_Rotate:TImageRotate;
begin
  result := false;
  BMP:=nil;
  ExifData:=TExif.create;
  if not Jpg2BMP(_JpgFilename,_Bitmap) then exit;
  ExifData.free;
  if not GetExifData(_JpgFilename,ExifData) then exit;
  ExifData.ImageWidth:=_Bitmap.Width;
  ExifData.ImageHeight:=_Bitmap.Height;
  _Rotate:=ExifOrientationToRotate(Byte(ExifData.Orientation));
  BMP:=TBitmap.Create;
  try
    try
      SmoothResize(_Bitmap,BMP,_NewWidth,_NewHeight);
      RotateBitmap(BMP,_Bitmap,_Rotate);
      BMP.Assign(_Bitmap);
      result:=true;
    except
      on e:exception do showmessage(e.Message);
    end;
  finally
    _Bitmap.Free;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: RotateBitmap
  Author:    sam
  Date:      31-Mai-2021
  Arguments: _SrcImage,_TrgImage:TBitmap;const _ImageRotate:TImageRotate
  Result:    None
-----------------------------------------------------------------------------}
procedure RotateBitmap(_SrcImage,_TrgImage:TBitmap;const _ImageRotate:TImageRotate);
var
  x,y:integer;
  p:array[0..2] of TPoint;
begin
  x:=_SrcImage.Width;
  y:=_SrcImage.Height;

  case _ImageRotate of
    tr0: begin
      _TrgImage.Assign(_SrcImage)
    end;
    tr90: begin
      _TrgImage.Width:=y;
      _TrgImage.Height:=x;
      p[0].X:=y;
      p[0].Y:=0;
      p[1].X:=y;
      p[1].Y:=x;
      p[2].X:=0;
      p[2].Y:=0;
    end;
    tr180:begin
      _TrgImage.Width:=x;
      _TrgImage.Height:=y;
      p[0].X:=0;
      p[0].Y:=y;
      p[1].X:=x;
      p[1].Y:=y;
      p[2].X:=0;
      p[2].Y:=0;
    end;
    tr270:begin
      _TrgImage.Width:=y;
      _TrgImage.Height:=x;
      p[0].X:=0;
      p[0].Y:=x;
      p[1].X:=0;
      p[1].Y:=0;
      p[2].X:=y;
      p[2].Y:=x;
    end;
  end;
  if _ImageRotate<>tr0 then
   {$IFDEF FPC}
    // ???
   {$ELSE}
    PlgBlt(_TrgImage.Canvas.Handle,p,_SrcImage.Canvas.Handle,0,0,x,y,0,0,0);
   {$ENDIF}
end;

{-----------------------------------------------------------------------------
  Procedure: DetermBackgroundColor
  Author:    sam
  Date:      31-Mai-2021
  Arguments: const _Bitmap:TBitmap;out red_min,red_max,green_min,green_max,blue_min,blue_max:byte
  Result:    None
-----------------------------------------------------------------------------}
function DetermBackgroundColor(const _Bitmap:TBitmap;out red_min,red_max,green_min,green_max,blue_min,blue_max:byte):boolean;
var
y,x: integer;
_mx,_my:integer;
row: ^TRGBTriple;
begin
  result:=false;
  if not assigned(_Bitmap) then exit;

  red_min:=255;
  red_max:=0;
  green_min:=255;
  green_max:=0;
  blue_min:=255;
  blue_max:=0;

  _mx:=_Bitmap.Width div 2;
  _my:=_Bitmap.Height div 2;
  for y:= 0 to _Bitmap.Height-1 do begin
    row:= _Bitmap.Scanline[y];
    for x:= 0 to _Bitmap.Width-1 do begin
      if ((x=1) and (y=1)) or
         ((x=_mx) and (y=1)) or
         ((x=_Bitmap.Width-2) and (y=1)) or
         ((x=_Bitmap.Width-2) and (y=_my)) or
         ((x=_Bitmap.Width-2) and (y=_Bitmap.Height-2)) or
         ((x=_mx) and (y=_Bitmap.Height-2)) or
         ((x=1) and (y=_Bitmap.Height-2)) or
         ((x=1) and (y=_my)) then begin
        if row^.rgbtred<red_min then red_min:= row^.rgbtred;
        if row^.rgbtred>red_max then red_max:= row^.rgbtred;
        if row^.rgbtgreen<green_min then green_min:= row^.rgbtgreen;
        if row^.rgbtgreen>green_max then green_max:= row^.rgbtgreen;
        if row^.rgbtblue<blue_min then blue_min:= row^.rgbtblue;
        if row^.rgbtblue>blue_max then blue_max:= row^.rgbtblue;
      end;
      inc(row);
    end;
  end;
  result:=true;
end;

{-----------------------------------------------------------------------------
  Procedure: DetermBackgroundValue
  Author:    sam
  Date:      20-Nov-2017
  Arguments: const _Bitmap:TBitmap; out backgroundValue: byte
  Result:    None
  Description: auto detect "background"-value from 8 Pixels around the boarder
               Take the red-value and calculate the average.
-----------------------------------------------------------------------------}
procedure DetermBackgroundValue(const _Bitmap:TBitmap; out BackgroundValue: byte);
var
i:integer;
_Pixel:TColor;
_Pixels:array[0..7] of integer;
_sum:integer;
_sumvalue:double;
begin
  BackgroundValue:=0;
  _Pixel:=_Bitmap.Canvas.Pixels[_Bitmap.Width div 2,5];   // top middle
  _Pixels[0]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);
  _Pixel:=_Bitmap.Canvas.Pixels[5,5];    // top left
  _Pixels[1]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);
  _Pixel:=_Bitmap.Canvas.Pixels[_Bitmap.width-5,5];   // top right
  _Pixels[2]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);
  _Pixel:=_Bitmap.Canvas.Pixels[5,_Bitmap.Height div 2]; // left middle
  _Pixels[3]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);
  _Pixel:=_Bitmap.Canvas.Pixels[_Bitmap.width-5,_Bitmap.Height div 2]; // right middle
  _Pixels[4]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);
  _Pixel:=_Bitmap.Canvas.Pixels[5,_Bitmap.Height-5]; // bottom left
  _Pixels[5]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);
  _Pixel:=_Bitmap.Canvas.Pixels[_Bitmap.width-5,_Bitmap.Height-5]; // bottom right
  _Pixels[6]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);
  _Pixel:=_Bitmap.Canvas.Pixels[_Bitmap.width div 2,_Bitmap.Height-5]; // bottom middle
  _Pixels[7]:=trunc((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel))/3);

  _sum:=0;
  for i := 0 to 7 do _sum:=_sum+_Pixels[i];
  _sumvalue:=(_sum / 8);
  BackgroundValue:=round(_sumvalue)
end;


procedure DetermBackgroundValue(const _Bitmap:TBitmap;const _Exif:TExif; out BackgroundValue: byte);
begin
  BackgroundValue:=trunc(125+_Exif.Brightness);
end;

{-----------------------------------------------------------------------------
  Procedure: DetermForegroundValue
  Author:    sam
  Date:      28-Jul-2024
  Arguments: const _Bitmap:TBitmap; out foregroundValue: byte
  Result:    None
  Description: Take the red-value and calculate the average.
-----------------------------------------------------------------------------}
procedure DetermForegroundValue(const _Bitmap:TBitmap; out foregroundValue: byte);
var
i:integer;
_Pixel:TColor;
_Pixels:array[0..4] of integer;
_sum:integer;
_sumvalue:double;
begin
  foregroundValue:=0;
  _Pixel:=_Bitmap.Canvas.Pixels[(_Bitmap.Width div 2)-30, (_Bitmap.Height div 2)-30];
  _Pixels[0]:=round((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel)) / 3);
  _Pixel:=_Bitmap.Canvas.Pixels[(_Bitmap.Width div 2)+30, (_Bitmap.Height div 2)-30];
  _Pixels[1]:=round((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel)) / 3);
  _Pixel:=_Bitmap.Canvas.Pixels[(_Bitmap.Width div 2)-30, (_Bitmap.Height div 2)+30];
  _Pixels[2]:=round((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel)) / 3);
  _Pixel:=_Bitmap.Canvas.Pixels[(_Bitmap.Width div 2)+30, (_Bitmap.Height div 2)+30];
  _Pixels[3]:=round((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel)) / 3);
  _Pixel:=_Bitmap.Canvas.Pixels[_Bitmap.Width div 2,_Bitmap.Height div 2];
  _Pixels[4]:=round((GetRValue(_Pixel)+GetGValue(_Pixel)+GetBValue(_Pixel)) / 3);

  _sum:=0;
  for i := 0 to 4 do _sum:=_sum+_Pixels[i];
  _sumvalue:=(_sum / 5);

  if round(_sumvalue)<256 then foregroundValue:=round(_sumvalue)
                          else foregroundValue:=255;
end;

{-----------------------------------------------------------------------------
  Procedure: BlackWhiteImage
  Author:
  Date:
  Arguments: const Bitmap: TBitmap; const threshold: Integer
  Result:    None
  Description: make the image black&white
-----------------------------------------------------------------------------}
procedure BlackWhiteImage(var Bitmap:TBitmap; const _threshold: byte;out ObjectArea:TRect);
var
_y,_x: integer;
_row: ^TRGBTriple;
_NewValue:byte;
begin
  ObjectArea.Left:=65000;
  ObjectArea.Top:=65000;
  ObjectArea.Right:=-1;
  ObjectArea.Bottom:=-1;

  for _y:= 0 to Bitmap.height-1 do begin
    _row:= Bitmap.Scanline[_y];
    for _x:=0 to Bitmap.width-1 do begin

      if _row^.rgbtred>_threshold then _NewValue:=255
                                  else _NewValue:=0;

      if _row^.rgbtred<>_NewValue   then _row^.rgbtred   :=_NewValue;
      if _row^.rgbtBlue<>_NewValue  then _row^.rgbtBlue  :=_NewValue;
      if _row^.rgbtgreen<>_NewValue then _row^.rgbtgreen :=_NewValue;
      if _NewValue=255 then begin
        if _x<ObjectArea.Left   then ObjectArea.Left :=_x;
        if _x>ObjectArea.Right  then ObjectArea.Right:=_x;
        if _y<ObjectArea.Top    then ObjectArea.Top  :=_y;
        if _y>ObjectArea.Bottom then ObjectArea.Bottom:=_y;
      end;
      inc(_row);
    end;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: AutoBlackWhiteImage
  Author:    sam
  Date:      20-Apr-2019
  Arguments: var Bitmap: Vcl.Graphics.TBitmap
  Result:    boolean
-----------------------------------------------------------------------------}
function AutoBlackWhiteImage(var Bitmap: TBitmap;const _Exif:TExif;out ObjectArea:TRect):boolean;
var
_BackgroundValue:Byte;
begin
  result:=false;
  try
    DetermBackgroundValue(Bitmap,_Exif,_BackgroundValue);
    BlackWhiteImage(Bitmap,_BackgroundValue,ObjectArea);
    result:=true;
  except
    on e:exception do showmessagefmt('Error in AutoBlackWhiteImage: <%s>.',[e.message]);
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: ManipulateImage
  Author:    sam
  Date:      18-Mai-2021
  Arguments: var Bitmap:TBitmap
  Result:    None
  Description: manipulate image <Bitmap> with operation <_ImageOperation>.
-----------------------------------------------------------------------------}
procedure ManipulateImage(var Bitmap:TBitmap;const _ImageOperation:TImageOperation);
var
y,x: integer;
row: ^TRGBTriple;

function CalcAverage:byte;
begin
  result:=(row^.rgbtred+row^.rgbtBlue+row^.rgbtgreen) div 3;
  row^.rgbtred   := result;
  row^.rgbtBlue  := result;
  row^.rgbtgreen := result;
end;

function CalcMin:byte;
begin
  result:=row^.rgbtred;
  if row^.rgbtBlue<result   then result:=row^.rgbtBlue;
  if row^.rgbtgreen<result  then result:=row^.rgbtgreen;
  row^.rgbtred   := result;
  row^.rgbtBlue  := result;
  row^.rgbtgreen := result;
end;

function CalcMax:byte;
begin
  result:=row^.rgbtred;
  if row^.rgbtBlue>result   then result:=row^.rgbtBlue;
  if row^.rgbtgreen>result  then result:=row^.rgbtgreen;
  row^.rgbtred   := result;
  row^.rgbtBlue  := result;
  row^.rgbtgreen := result;
end;


procedure CalcMaxSubMin;
var
_l,_h:byte;
_cl,_ch:integer;
_d:byte;
begin
  _l:=255;
  _h:=0;
  _cl:=0;
  _ch:=0;
  if row^.rgbtRed>_h   then begin
    _h:=row^.rgbtRed;
    _ch:=1;
  end;
  if row^.rgbtGreen>_h then begin
    _h:=row^.rgbtGreen;
    _ch:=2;
  end;
  if row^.rgbtBlue>_h  then begin
    _ch:=3;
  end;

  if row^.rgbtRed<_l   then begin
    _l:=row^.rgbtRed;
    _cl:=1;
  end;
  if row^.rgbtGreen<_l then begin
    _l:=row^.rgbtGreen;
    _cl:=2;
  end;
  if row^.rgbtBlue<_l  then begin
    _cl:=3;
  end;
  _d:=0;
  case _cl of
    1:begin
      case _ch of
        1:;
        2: if row^.rgbtRed>row^.rgbtGreen then _d:=0
                                          else _d:=row^.rgbtGreen-row^.rgbtRed;
        3: if row^.rgbtRed>row^.rgbtBlue  then _d:=0
                                          else _d:=row^.rgbtBlue -row^.rgbtRed;
      end;
    end;
    2:begin
      case _ch of
        1: if row^.rgbtGreen>row^.rgbtRed then _d:=0
                                          else _d:=row^.rgbtRed-row^.rgbtGreen;
        2: ;
        3: if row^.rgbtGreen>row^.rgbtBlue then _d:=0
                                           else _d:=row^.rgbtBlue-row^.rgbtGreen;
      end;
    end;
    3:begin
      case _ch of
        1: if row^.rgbtBlue>row^.rgbtRed then _d:=0
                                         else _d:=row^.rgbtRed  -row^.rgbtBlue;
        2: if row^.rgbtBlue>row^.rgbtGreen then _d:=0
                                           else _d:=row^.rgbtGreen-row^.rgbtBlue;
        3: ;
      end;
    end;
  end;

  if row^.rgbtred +_d>255 then row^.rgbtred:=255
                          else row^.rgbtred:= row^.rgbtred +_d;
  if row^.rgbtBlue+_d>255 then row^.rgbtBlue:= 255
                          else row^.rgbtBlue  := row^.rgbtBlue+_d;
  if row^.rgbtgreen+_d>255 then row^.rgbtgreen := 255
                           else row^.rgbtgreen := row^.rgbtgreen+_d;
end;

procedure CalcMaxSubAverage;
var
_h:byte;
_ch:integer;
_d,_a:byte;
begin
  _h:=0;
  _ch:=0;
  _a:=round((row^.rgbtRed+row^.rgbtGreen+row^.rgbtBlue) / 3);
  if row^.rgbtRed>_h   then begin
    _h:=row^.rgbtRed;
    _ch:=1;
  end;
  if row^.rgbtGreen>_h then begin
    _h:=row^.rgbtGreen;
    _ch:=2;
  end;
  if row^.rgbtBlue>_h  then begin
    _ch:=3;
  end;

  _d:=0;
  case _ch of
    1: _d:= row^.rgbtRed   - _a;
    2: _d:= row^.rgbtGreen - _a;
    3: _d:= row^.rgbtBlue  - _a;
  end;

  row^.rgbtred   := _d;
  row^.rgbtBlue  := _d;
  row^.rgbtgreen := _d;
end;

function CalcInvert:byte;
begin
  result:=Abs(255-(row^.rgbtred+row^.rgbtBlue+row^.rgbtgreen) div 3);
  row^.rgbtred   := result;
  row^.rgbtBlue  := result;
  row^.rgbtgreen := result;
end;

function CalcBlackWhite:byte;
begin
  if ((row^.rgbtred+row^.rgbtBlue+row^.rgbtgreen) div 2)>80 then result:=255
                                                             else result:=0;
  if row^.rgbtred<>result   then row^.rgbtred   := result;
  if row^.rgbtBlue<>result  then row^.rgbtBlue  := result;
  if row^.rgbtgreen<>result then row^.rgbtgreen := result;
end;

function CalcNormalize:byte;
var
_Mean:double;
_pMean:double;
_v:double;
_r,_g,_b:double;
begin
  _r:=row^.rgbtred;
  _g:=row^.rgbtgreen;
  _b:=row^.rgbtBlue;
  _Mean:=(_r+_g+_b) / 3;
  _pMean:=((_r+_g+_b) / (3*255));

  if (_pMean<0.6) and
     (_pMean>0.4) then _pMean:=1.4

  else
  if (_pMean<0.8) and
     (_pMean>0.2) then _pMean:=2;

  if _Mean=0 then begin
    result:=0;
    exit;
  end;


  _v:= (_r *_pMean);
  if _v>255 then row^.rgbtred:=255
            else row^.rgbtred:=trunc(_v);

  _v:= (_g * _pMean);
  if _v>255 then row^.rgbtgreen:=255
            else row^.rgbtgreen:=trunc(_v);

  _v:= (_b *_pMean);
  if _v>255 then row^.rgbtblue:=255
            else row^.rgbtblue:=trunc(_v);
  result:= ((row^.rgbtred+row^.rgbtgreen+row^.rgbtblue) div 3);
end;

function CalcPushRed:byte;
var
_sum:byte;
_v:double;
begin
  _sum:=(row^.rgbtred+row^.rgbtBlue+row^.rgbtgreen) div 3;
  if _sum=0 then begin
    result:=0;
    exit;
  end;
  if _sum>130 then begin
    if row^.rgbtred>_sum then begin
      _v:=row^.rgbtred*1.2;
      if _v>255 then row^.rgbtred:=255
                else row^.rgbtred:=trunc(_v);
    end;
    row^.rgbtgreen:=_sum;
    row^.rgbtblue:=_sum;
  end
  else begin
    row^.rgbtred:=0;
    row^.rgbtgreen:=0;
    row^.rgbtblue:=0;
  end;
//  if row^.rgbtgreen<_sum then row^.rgbtgreen:=round(_sum*1);
//  if row^.rgbtblue<_sum then row^.rgbtblue:=round(_sum*1);
  result:= ((row^.rgbtred+row^.rgbtgreen+row^.rgbtblue) div 3);
end;

procedure SetColor(const _r,_g,_b:integer);
begin
  if (_r>-1) and (row^.rgbtred<>_r)   and (_r<256) then row^.rgbtred:=_r;
  if (_g>-1) and (row^.rgbtgreen<>_g) and (_g<256) then row^.rgbtgreen:=_g;
  if (_b>-1) and (row^.rgbtblue<>_b)  and (_b<256) then row^.rgbtblue:=_b;
end;


procedure CalcChangeBrightness(const _value:integer);
var
_NewValue:integer;
begin
  if row^.rgbtblue>50 then begin
    _NewValue:=row^.rgbtblue+_value;
    if (_NewValue<0)   then row^.rgbtblue:=0 else
    if (_NewValue>255) then row^.rgbtblue:=255 else
    row^.rgbtblue:=_NewValue;
  end;

  if row^.rgbtred>50 then begin
    _NewValue:=row^.rgbtred+_value;
    if (_NewValue<0)   then row^.rgbtred:=0 else
    if (_NewValue>255) then row^.rgbtred:=255 else
    row^.rgbtred:=_NewValue;
  end;

  if row^.rgbtgreen>50 then begin
    _NewValue:=row^.rgbtgreen+_value;
    if (_NewValue<0)   then row^.rgbtgreen:=0 else
    if (_NewValue>255) then row^.rgbtgreen:=255 else
    row^.rgbtgreen:=_NewValue;
  end;
end;

begin
  for y:= 0 to Bitmap.Height-1 do begin
    row:= Bitmap.Scanline[y];
    for x:= 0 to Bitmap.Width-1 do begin
      case _ImageOperation of
        tgs_average: CalcAverage;
        tgs_min    : CalcMin;
        tgs_max    : CalcMax;
        tgs_maxsubmin: CalcMaxSubMin;
        tgs_maxsubavarage:CalcMaxSubAverage;
        tgs_invert:CalcInvert;
        tgs_blackwhite:CalcBlackWhite;
        tgs_normalize:CalcNormalize;
        tgs_push_red:CalcPushRed;
        tgs_remove_blue:  SetColor(-1,-1,0);
        tgs_remove_green: SetColor(-1,0,-1);
        tgs_keep_red_only:SetColor(-1,-0,0);
        tgs_increase_brightness:CalcChangeBrightness(50);
      end;
      inc(row);
    end;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: Crop
  Author:    sam
  Date:      08-Jun-2022
  Arguments: var Bitmap: TBitmap;const _SourceRect:TRect
  Result:    boolean
  Description: Copy bitmap <Bitmap> to rect <_SourceRect>.
-----------------------------------------------------------------------------}
function Crop(var Bitmap: TBitmap;const _SourceRect:TRect):boolean;
var
_DestRect: TRect;
_OriginalBmp : TBitmap;
begin
  result:=false;
  if not assigned(Bitmap) then exit;

  _OriginalBmp := TBitmap.Create;
  try
    _OriginalBmp.Assign(Bitmap);
    Bitmap.free;
    Bitmap:= TBitmap.Create;
    Bitmap.Width  := _SourceRect.Width;
    Bitmap.Height := _SourceRect.height;
    Bitmap.PixelFormat:=pf24bit;
    _DestRect.Left:=0;
    _DestRect.Top:=0;
    _DestRect.Right:=Bitmap.Width;
    _DestRect.Bottom:=Bitmap.Height;
    Bitmap.Canvas.CopyRect(_DestRect,_OriginalBmp.Canvas,_SourceRect);
    result:=true;
  finally
    _OriginalBmp.free;
  end;
end;

{-----------------------------------------------------------------------------
  Procedure: SmoothResize
  Author:    swissdelphicenter
  Date:      04-Jul-2020
  Arguments: Src, Dst: TBitmap
  Result:    None
-----------------------------------------------------------------------------}
procedure SmoothResize(Src, Dst: TBitmap;_NewWidth,_NewHeight:integer);
var
x, y: Integer;
xP, yP: Integer;
xP2, yP2: Integer;
SrcLine1, SrcLine2: pRGBArray;
t3: Integer;
z, z2, iz2: Integer;
DstLine: pRGBArray;
DstGap: Integer;
w1, w2, w3, w4: Integer;
begin

  if (_NewWidth=0) and
     (_NewHeight=0) then begin
    Dst.Assign(Src);
    exit;
  end;


  if _NewWidth=0 then begin
    if _NewHeight>0 then _NewWidth:=trunc(Src.Width*_NewHeight/Src.Height)
                    else _NewWidth:=Src.Width;
  end;
  Dst.Width :=_NewWidth;
  if _NewHeight=0 then begin
    if _NewWidth>0 then _NewHeight:=trunc(Src.Height*_NewWidth/Src.Width)
                   else _NewHeight:=Src.Height;
  end;
  Dst.Height:=_NewHeight;


  Src.PixelFormat := pf24Bit;
  Dst.PixelFormat := pf24Bit;

  if (Src.Width = Dst.Width) and (Src.Height = Dst.Height) then begin
    Dst.Assign(Src);
    exit;
  end;
  DstLine := Dst.ScanLine[0];
  DstGap  := Integer(Dst.ScanLine[1]) - Integer(DstLine);

  xP2 := MulDiv(pred(Src.Width) , $10000, Dst.Width);
  yP2 := MulDiv(pred(Src.Height), $10000, Dst.Height);
  yP  := 0;

  for y := 0 to pred(Dst.Height) do
  begin
    xP := 0;

    SrcLine1 := Src.ScanLine[yP shr 16];

    if (yP shr 16 < pred(Src.Height)) then
      SrcLine2 := Src.ScanLine[succ(yP shr 16)]
    else
      SrcLine2 := Src.ScanLine[yP shr 16];

    z2  := succ(yP and $FFFF);
    iz2 := succ((not yp) and $FFFF);
    for x := 0 to pred(Dst.Width) do
    begin
      t3 := xP shr 16;
      z  := xP and $FFFF;
      w2 := MulDiv(z, iz2, $10000);
      w1 := iz2 - w2;
      w4 := MulDiv(z, z2, $10000);
      w3 := z2 - w4;
      DstLine[x].rgbtRed := (SrcLine1[t3].rgbtRed * w1 +
        SrcLine1[t3 + 1].rgbtRed * w2 +
        SrcLine2[t3].rgbtRed * w3 + SrcLine2[t3 + 1].rgbtRed * w4) shr 16;
      DstLine[x].rgbtGreen :=
        (SrcLine1[t3].rgbtGreen * w1 + SrcLine1[t3 + 1].rgbtGreen * w2 +

        SrcLine2[t3].rgbtGreen * w3 + SrcLine2[t3 + 1].rgbtGreen * w4) shr 16;
      DstLine[x].rgbtBlue := (SrcLine1[t3].rgbtBlue * w1 +
        SrcLine1[t3 + 1].rgbtBlue * w2 +
        SrcLine2[t3].rgbtBlue * w3 +
        SrcLine2[t3 + 1].rgbtBlue * w4) shr 16;
      Inc(xP, xP2);
    end;
    Inc(yP, yP2);
    DstLine := pRGBArray(Integer(DstLine) + DstGap);
  end;
//  Dst.SaveToFile('c:\temp\resized.bmp');
end;

function ConvertToGrayscale(const _Source: TBitmap;out Target:TBitmap): boolean;
var
  X, Y: Integer;
  GrayColor: Byte;
begin
  result:=false;
  if not assigned(_Source) then exit;
  Target := TBitmap.Create;
  Target.Assign(_Source);
  Target.PixelFormat := pf24bit;
  for Y := 0 to Target.Height - 1 do begin
    for X := 0 to Target.Width - 1 do begin
      GrayColor := Round(0.299 * GetRValue(Target.Canvas.Pixels[X, Y]) +
                         0.587 * GetGValue(Target.Canvas.Pixels[X, Y]) +
                         0.114 * GetBValue(Target.Canvas.Pixels[X, Y]));
      Target.Canvas.Pixels[X, Y] := RGB(GrayColor, GrayColor, GrayColor);
    end;
  end;
  result:=true;
end;

function ResizeBitmap(const _Source: TBitmap; const _Width, _Height: Integer;out Target:TBitmap): boolean;
begin
  result:=false;
  if not assigned(_Source) then exit;
  Target := TBitmap.Create;
  Target.SetSize(_Width, _Height);
  Target.Canvas.StretchDraw(Rect(0, 0, _Width, _Height), _Source);
  result:=true;
end;


end.
